#!/usr/bin/env python
import os
import re
import sys
import time
from collections import Iterable
from collections import OrderedDict

if sys.version_info[0] == 2:
    from urllib2 import URLError
else:
    from urllib.error import URLError

import xmltodict
import numpy as np
from Bio import SeqIO, Entrez
from Bio.SeqRecord import SeqRecord
from RouToolPa.Routines.Sequence import SequenceRoutines
from RouToolPa.Routines.Annotations import AnnotationsRoutines
from RouToolPa.Parsers.GFF import CollectionGFF
from RouToolPa.Collections.General import IdList, SynDict, TwoLvlDict, IdSet


class AssemblySummary(OrderedDict):

    def __init__(self, assembly_summary_biopython_dict_element):
        OrderedDict.__init__(self)
        #parameters_dict = entrez_summary_biopython['DocumentSummarySet']['DocumentSummary'][0]
        self.stats = self.parse_stats_from_assembly_summary_metadata(assembly_summary_biopython_dict_element['Meta'])

        for parameter in 'SpeciesName', 'Organism', 'Taxid', 'AssemblyName', 'FtpPath_GenBank', 'PropertyList', \
                         'SubmissionDate', 'LastUpdateDate', 'AssemblyAccession', 'LastMajorReleaseAccession', \
                         'AssemblyType', 'PartialGenomeRepresentation', 'AssemblyClass', 'AnomalousList',\
                         'AssemblyStatus', 'BioSampleId', 'BioSampleAccn', 'WGS', 'Biosource':
            if parameter in assembly_summary_biopython_dict_element:
                setattr(self, parameter, assembly_summary_biopython_dict_element[parameter])
            else:
                setattr(self, parameter, None)
                #exec("self[%s] = None" % parameter)

        if self.Biosource:
            #print self.Biosource
            self['Isolate'] = self.Biosource['Isolate'] if self.Biosource['Isolate'] else None
            if ('InfraspeciesList' in self.Biosource) and self.Biosource['InfraspeciesList']:
                self['Sub_type'] = self.Biosource['InfraspeciesList'][0]['Sub_type'] if 'Sub_type' in self.Biosource['InfraspeciesList'][0] else None
                self['Sub_value'] = self.Biosource['InfraspeciesList'][0]['Sub_value'] if 'Sub_value' in self.Biosource['InfraspeciesList'][0] else None
            else:
                self['Sub_type'] = None
                self['Sub_value'] = None
            if ('Sex' in self.Biosource) and self.Biosource['Sex']:
                self['Sex'] = self.Biosource['Sex']
            else:
                self['Sex'] = None

        else:
            self['Isolate'] = None
            self['Sub_type'] = None
            self['Sub_value'] = None
            self['Sex'] = None

        for parameter in 'Isolate', 'Sub_type', 'Sub_value':
            setattr(self, parameter, self[parameter])

        for entry in self.stats:
            self[entry] = self.stats[entry]
        for entry in assembly_summary_biopython_dict_element:
            self[entry] = assembly_summary_biopython_dict_element[entry]

        for parameter in "alt_loci_count", "chromosome_count", "contig_count", "contig_l50", "contig_n50", \
                         "non_chromosome_replicon_count", "replicon_count", "scaffold_count_all", \
                         "scaffold_count_placed", "scaffold_count_unlocalized", "scaffold_count_unplaced",\
                         "scaffold_l50", "scaffold_n50", "total_length", "ungapped_length":
            setattr(self, parameter, self[parameter])

        self.full_genome = True if 'full-genome-representation' in self.PropertyList else False
        self.chromosome_lvl = True if 'has-chromosome' in self.PropertyList else False
        self.chloroplast = True if 'has-chloroplast' in self.PropertyList else False
        self.mitochondrion = True if 'has-mitochondrion' in self.PropertyList else False
        self.plasmid = True if 'has-plasmid' in self.PropertyList else False

        self.header_list = ['SpeciesName', 'Taxid', 'Isolate', 'Sub_type', 'Sub_value', 'Sex', 'BioSampleId', 'BioSampleAccn',
                            'WGS', 'AssemblyAccession', 'AssemblyName', 'SubmissionDate', 'LastUpdateDate',
                            'AssemblyType', 'AssemblyStatus',
                            'WholeGenome', "ChromosomeLvl", 'Chloroplast', "Mitochondrion", "Plasmid",
                            'total_length', 'ungapped_length', "alt_loci_count",
                            'chromosome_count', 'non_chromosome_replicon_count', 'replicon_count',
                            'scaffold_count_all', 'scaffold_count_placed', 'scaffold_count_unlocalized',
                            'scaffold_count_unplaced', 'scaffold_l50', 'scaffold_n50',
                            'contig_count', 'contig_l50', 'contig_n50', 'FtpPath_GenBank']
        self.is_hybrid = True if " x " in self.SpeciesName else False
        self.unknown = True if ("unknown" in self.SpeciesName) or ("Unknown" in self.SpeciesName) else False
        self.only_genus_known = True if (" sp. " in self.SpeciesName) else False
        self.candidate_species = True if ("candidate" in self.SpeciesName) or ("Candidate" in self.SpeciesName) else False


    @staticmethod
    def parse_stats_from_assembly_summary_metadata(metadata):

        end = re.search("/Stats>", metadata).end()
        tmp_dict = xmltodict.parse(metadata[:end])
        stat_dict = OrderedDict()
        for entry in tmp_dict["Stats"]["Stat"]:
            if "scaffold_count" in entry['@category']:
                stat_dict[entry['@category'] + "_" + entry['@sequence_tag']] = int(entry['#text'])
            else:
                stat_dict[entry['@category']] = int(entry['#text'])

        return stat_dict

    def get_header(self):
        return "\t".join(self.header_list)

    def __str__(self, separator="\t"):
        values_list = []
        for key in self.header_list:
            if key == 'WholeGenome':
                values_list.append("Y" if self.full_genome else "N")
            elif key == "ChromosomeLvl":
                values_list.append("Y" if self.chromosome_lvl else "N")
            elif key == "Chloroplast":
                values_list.append("Y" if self.chloroplast else "N")
            elif key == "Mitochondrion":
                values_list.append("Y" if self.chloroplast else "N")
            elif key == "Plasmid":
                values_list.append("Y" if self.plasmid else "N")
            else:
                if isinstance(self[key], str):
                    tmp = (self[key])
                elif isinstance(self[key], list) or isinstance(self[key], set):
                    tmp = ','.join(self[key])
                elif self[key] is None:
                    tmp = "."
                else:
                    tmp = str(self[key])
                values_list.append(tmp)
        #print values_list
        return separator.join(values_list)

    def ambiguous_species(self):
        return True if (self.is_hybrid or self.unknown or self.only_genus_known or self.candidate_species) else False

    def check_integrity(self, min_scaffold_n50=None, min_contig_n50=None, max_scaffold_l50=None,
                        max_contig_l50=None, max_contig_count=None, max_scaffold_count=None,
                        max_chromosome_count=None, min_chromosome_count=None, max_unlocalized_scaffolds=None,
                        max_unplaced_scaffolds=None, max_total_length=None, min_total_length=None,
                        max_ungapped_length=None, min_ungapped_length=None):
        if min_contig_n50:
            if self.contig_n50 < min_contig_n50:
                return False
        if min_scaffold_n50:
            if self.scaffold_n50 < min_scaffold_n50:
                return False
        if max_scaffold_l50:
            if self.scaffold_l50 > max_scaffold_l50:
                return False
        if max_contig_l50:
            if self.contig_l50 > max_contig_l50:
                return False
        if max_contig_count:
            if self.contig_count > max_contig_count:
                return False
        if max_scaffold_count:
            if self.scaffold_count_all > max_scaffold_count:
                return False
        if max_chromosome_count:
            if self.chromosome_count > max_chromosome_count:
                return False
        if min_chromosome_count:
            if self.chromosome_count < min_chromosome_count:
                return False
        if max_unlocalized_scaffolds:
            if self.scaffold_count_unlocalized > max_unlocalized_scaffolds:
                return False
        if max_unplaced_scaffolds:
            if self.scaffold_count_unplaced < max_unplaced_scaffolds:
                return False
        if max_total_length:
            if self.total_length > max_total_length:
                return False
        if min_total_length:
            if self.total_length < min_total_length:
                return False

        if max_ungapped_length:
            if self.ungapped_length > max_ungapped_length:
                return False
        if min_ungapped_length:
            if self.ungapped_length < min_ungapped_length:
                return False

        return True


class AssemblySummaryList(list):
    def __init__(self, entrez_summary_biopython=None):
        list.__init__(self)
        if entrez_summary_biopython:
            for assembly_summary_biopython_dict_element in entrez_summary_biopython['DocumentSummarySet']['DocumentSummary']:
            #print assembly_summary_biopython_dict_element
                self.append(AssemblySummary(assembly_summary_biopython_dict_element))

            self.header = self[0].get_header()

        else:
            self.header = "\t".join(['SpeciesName', 'Taxid', 'Isolate', 'Sub_type', 'Sub_value', 'Sex', 'BioSampleId',
                                     'BioSampleAccn', 'WGS', 'AssemblyAccession', 'AssemblyName', 'SubmissionDate',
                                     'LastUpdateDate', 'AssemblyType', 'AssemblyStatus',
                                     'WholeGenome', "ChromosomeLvl", 'Chloroplast', "Mitochondrion", "Plasmid",
                                     'total_length', 'ungapped_length', "alt_loci_count",
                                     'chromosome_count', 'non_chromosome_replicon_count', 'replicon_count',
                                     'scaffold_count_all', 'scaffold_count_placed', 'scaffold_count_unlocalized',
                                     'scaffold_count_unplaced', 'scaffold_l50', 'scaffold_n50',
                                     'contig_count', 'contig_l50', 'contig_n50', 'FtpPath_GenBank'])

    def __str__(self):
        string = self.header + "\n"
        for summary in self:
            string += str(summary) + "\n"

    def write(self, output_file):
        with open(output_file, "w") as out_fd:
            out_fd.write(self.header + "\n")
            for summary in self:
                out_fd.write(str(summary) + "\n")

    def filter(self, expression):
        filtered = AssemblySummaryList()
        filtered_out = AssemblySummaryList()

        for summary in self:
            if expression(summary):
                filtered.append(summary)
            else:
                filtered_out.append(summary)

        return filtered, filtered_out

    def filter_non_chrom_level_genomes(self):
        def chromosome_lvl(summary):
            return summary.chromosome_lvl

        return self.filter(chromosome_lvl)

    def filter_ambiguous_species(self, verbose=False):
        if verbose:
            for summary in self:
                print("\tSpeciesName\tIsAmbigious")
                print("\t%s\t%s" % (str(summary.SpeciesName), str(summary.ambiguous_species())))
        return self.filter(lambda summary: not summary.ambiguous_species())

    def filter_by_integrity(self, min_scaffold_n50=None, min_contig_n50=None, max_scaffold_l50=None,
                            max_contig_l50=None, max_contig_count=None, max_scaffold_count=None,
                            max_chromosome_count=None, min_chromosome_count=None, max_unlocalized_scaffolds=None,
                            max_unplaced_scaffolds=None, max_total_length=None, min_total_length=None,
                            max_ungapped_length=None, min_ungapped_length=None,
                            no_ambiguous_species=False):

        def expression(summary):
            return summary.check_integrity(min_scaffold_n50=min_scaffold_n50, min_contig_n50=min_contig_n50,
                                           max_scaffold_l50=max_scaffold_l50, max_contig_l50=max_contig_l50,
                                           max_contig_count=max_contig_count, max_scaffold_count=max_scaffold_count,
                                           max_chromosome_count=max_chromosome_count,
                                           min_chromosome_count=min_chromosome_count,
                                           max_unlocalized_scaffolds=max_unlocalized_scaffolds,
                                           max_unplaced_scaffolds=max_unplaced_scaffolds,
                                           max_total_length=max_total_length, min_total_length=min_total_length,
                                           max_ungapped_length=max_ungapped_length,
                                           min_ungapped_length=min_ungapped_length) \
                   and ((not summary.ambiguous_species()) if no_ambiguous_species else True)

        return self.filter(expression)


class NCBIRoutines(SequenceRoutines):
    def __init__(self):
        SequenceRoutines.__init__(self)
        self.ncbi_ftp = "ftp://ftp-trace.ncbi.nlm.nih.gov/"

    def get_sra_ftp_path_from_id(self, sra_id):

        sra_reads_dir = "sra/sra-instant/reads/"
        id_code_dict = {
            "DRR": "ByRun",
            "ERR": "ByRun",
            "SRR": "ByRun",

            "DRX": "ByExp",
            "ERX": "ByExp",
            "SRX": "ByExp",

            "DRS": "BySample",
            "ERS": "BySample",
            "SRS": "BySample",

            "DRP": "ByStudy",
            "ERP": "ByStudy",
            "SRP": "ByStudy"
        }
        id_group = sra_id[:3]
        id_subgroup = sra_id[:6]

        id_type = id_code_dict[id_group]

        return "%s%s%s/sra/%s/%s/%s/%s.sra" % (self.ncbi_ftp, sra_reads_dir, id_type, id_group,
                                               id_subgroup, sra_id, sra_id)

    @staticmethod
    def efetch(database, id_list, out_file, retmode=None, rettype=None, seq_start=None, seq_stop=None, strand=None, verbose=False,
               number_of_retries=100, retry_delay=None, log_file="efetch.log"):
        # replacement for Biopython Entrez.efetch
        # Biopython Entrez.efetch is bugged - it ignores seq_start and seq_stop values
        # https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=669632474&retmode=text&rettype=gb&seq_start=10832751&seq_stop=10848091&strand=1
        query = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"

        query += "db=%s" % database

        query += "&id=%s" % (id_list if isinstance(id_list, str) else ",".join(id_list))
        query += "&retmode=%s" % retmode if retmode else ""
        query += "&rettype=%s" % rettype if rettype else ""

        query += "&seq_start=%s" % str(seq_start) if seq_start else ""
        query += "&seq_stop=%s" % str(seq_stop) if seq_stop else ""
        query += "&strand=%s" % str(strand) if strand else ""

        curl_options = " --retry %i" % number_of_retries if number_of_retries else ""
        curl_options += " --retry-delay %i" % retry_delay if retry_delay else ""
        curl_options += " '%s'" % query
        curl_options += " > %s" % out_file

        curl_string = "curl %s" % curl_options
        with open(log_file, "a") as log_fd:
            log_fd.write(curl_string + "\n")
        if verbose:
            print(curl_string)
        os.system(curl_string)

    def get_gene_sequences(self, email, query, retmax=100000, output_directory=None):
        if output_directory:
            self.safe_mkdir(output_directory)
        Entrez.email = email
        handle = Entrez.esearch(term=query, db="gene", rettype="xml", retmax=retmax)
        records = Entrez.read(handle, validate=False)
        gene_ids = records['IdList']
        for gene_id in gene_ids:
            print("Extracting sequence for gi %s" % gene_id)
            id_handle = Entrez.efetch("gene", id=gene_id, rettype="xml", retmax=retmax)
            record = Entrez.read(id_handle, validate=False)

            #print(record[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"])
            if "Entrezgene_locus" not in record[0]:
                continue
            if "Gene-commentary_seqs" not in record[0]["Entrezgene_locus"][0]:
                continue
            if "Seq-loc_int" not in record[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]:
                continue
            gi = record[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"]['Seq-interval_id']['Seq-id']['Seq-id_gi']
            start = int(record[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"]['Seq-interval_from'])
            end = int(record[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"]['Seq-interval_to']) + 1
            strand = record[0]["Entrezgene_locus"][0]["Gene-commentary_seqs"][0]["Seq-loc_int"]["Seq-interval"]['Seq-interval_strand']['Na-strand'].attributes['value']
            strand = 1 if strand == "plus" else 2
            # print(gi)
            # print(start, end, strand)
            # Biopython Entrez.efetch is bugged - it ignores seq_start and seq_stop values
            # eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=669632474&retmode=text&rettype=gb&seq_start=10832751&seq_stop=10848091&strand=1

            out_file = "%s/%s.gb" % (output_directory, gi)

            self.efetch("nuccore", gi, out_file, retmode="text", rettype="gb", seq_start=start, seq_stop=end,
                        strand=strand)
            time.sleep(0.4)

    @staticmethod
    def get_longest_proteins_from_protein_tab_file(protein_tab_file, output_prefix):
        sorted_tab_file = "%s.sorted.tab" % output_prefix
        longest_pep_tab_file = "%s.longest_pep.tab" % output_prefix
        longest_pep_id_file = "%s.longest_pep.ids" % output_prefix

        with open(protein_tab_file)as tmp_fd:
            tmp = tmp_fd.readline().strip().split("\t")
            gene_id_index = tmp.index("GeneID")
            protein_id_index = tmp.index("Protein product")
            length_index = tmp.index("Length")

        sort_string = "(head -n 1 %s && tail -n +2 %s | sort -k%i,%i -k%i,%inr) > %s" % (protein_tab_file,
                                                                                         protein_tab_file,
                                                                                         gene_id_index + 1,
                                                                                         gene_id_index + 1,
                                                                                         length_index + 1,
                                                                                         length_index + 1,
                                                                                         sorted_tab_file)

        os.system(sort_string)

        longest_pep_ids = IdSet()

        with open(sorted_tab_file, "r") as in_fd:
            with open(longest_pep_tab_file, "w") as out_fd:
                out_fd.write(in_fd.readline())
                prev_gene = ""
                for line in in_fd:
                    tmp = line.split("\t")
                    if tmp[gene_id_index] == prev_gene:
                        continue
                    prev_gene = tmp[gene_id_index]
                    out_fd.write(line)
                    longest_pep_ids.add(tmp[protein_id_index])

        longest_pep_ids.write(longest_pep_id_file)

    def get_cds_for_proteins(self, protein_id_list, output_prefix, download_chunk_size=100, temp_dir_prefix="temp",
                             keyword_for_mitochondrial_pep="mitochondrion"):
        duplicated_protein_id_list = IdList()
        protein_id_uniq_list = IdList()

        prev_protein_id = ""
        for protein_id in sorted(protein_id_list):
            if protein_id != prev_protein_id:
                protein_id_uniq_list.append(protein_id)
                prev_protein_id = protein_id

            else:
                duplicated_protein_id_list.append(protein_id)

        duplicated_protein_id_list.write("%s.duplicated.pep.ids" % output_prefix)
        if duplicated_protein_id_list:
            print("WARNING!!! Duplicated protein ids were detected...")
            with self.metaopen("%s.warnings" % output_prefix, "w") as warn_fd:
                warn_fd.write("WARNING!!! Duplicated protein ids were detected...")

        from RouToolPa.Tools.Abstract import Tool
        output_directory = self.check_dir_path(self.split_filename(output_prefix)[0])
        transcript_temp_dir = "%s%s_transcripts" % (output_directory, temp_dir_prefix)
        protein_temp_dir = "%s%s_proteins" % (output_directory, temp_dir_prefix)
        number_of_ids = len(protein_id_uniq_list)

        print("Total %i ids" % number_of_ids)

        for directory in transcript_temp_dir, protein_temp_dir:
            print(directory)
            self.safe_mkdir(directory)
        pep_file = "%s.pep.genbank" % output_prefix
        transcript_file = "%s.trascript.genbank" % output_prefix
        mito_transcript_file = "%s.trascript.mito.genbank" % output_prefix

        ranges = np.append(np.arange(0, number_of_ids, download_chunk_size), [number_of_ids])

        print("Downloading proteins...")
        for i in range(0, len(ranges)-1):
            print("Downloading chunk %i" % i)
            pep_tmp_file = "%s/chunk_%i" % (protein_temp_dir, i)
            self.efetch("protein", protein_id_uniq_list[ranges[i]:ranges[i+1]], pep_tmp_file, rettype="gb", retmode="text",
                        log_file="%s.pep.efetch.log" % output_prefix)

        os.system("cat %s/* > %s" % (protein_temp_dir, pep_file))

        peptide_dict = SeqIO.index_db("%stmp.idx" % output_directory, pep_file, format="genbank")
        downloaded_protein_ids = IdList(peptide_dict.keys())

        print("%i proteins were downloaded" % len(downloaded_protein_ids))
        not_downloaded_proteins_ids = Tool.intersect_ids([protein_id_uniq_list], [downloaded_protein_ids], mode="only_a")
        print("%i proteins were not downloaded" % len(not_downloaded_proteins_ids))
        not_downloaded_proteins_ids.write("%s.not_downloaded.ids" % output_prefix)
        downloaded_protein_ids.write("%s.downloaded.ids" % output_prefix)
        print(Tool.intersect_ids([protein_id_uniq_list], [downloaded_protein_ids], mode="count"))

        pep_without_transcripts = IdList()
        mito_pep_without_transcripts = IdList()
        pep_with_several_CDS_features = IdList()
        mito_pep_with_several_CDS_features = IdList()
        pep_to_transcript_accordance = SynDict()
        mito_pep_to_transcript_accordance = SynDict()
        transcript_ids = IdList()
        mito_transcript_ids = IdList()
        print("Extracting transcript ids corresponding to proteins...")

        for pep_id in peptide_dict:
            if keyword_for_mitochondrial_pep in peptide_dict[pep_id].annotations["source"]:
                for feature in peptide_dict[pep_id].features:
                    if feature.type == "CDS":
                        try:
                            transcript_id = feature.qualifiers["coded_by"][0].split(":")[0]
                            if pep_id not in mito_pep_to_transcript_accordance:
                                mito_pep_to_transcript_accordance[pep_id] = [transcript_id]
                            else:
                                mito_pep_to_transcript_accordance[pep_id].append(transcript_id)
                                print("Genbank record for %s contains several CDS features" % pep_id)
                                mito_pep_with_several_CDS_features.append(pep_id)
                            if transcript_id in mito_transcript_ids:
                                print("Repeated transcript id: %s" % transcript_id)
                                continue
                            mito_transcript_ids.append(transcript_id)
                        except:
                            print("Transcript id for %s was not found" % pep_id)
                            mito_pep_without_transcripts.append(pep_id)

            else:
                for feature in peptide_dict[pep_id].features:
                    if feature.type == "CDS":
                        try:
                            transcript_id = feature.qualifiers["coded_by"][0].split(":")[0]
                            if pep_id not in pep_to_transcript_accordance:
                                pep_to_transcript_accordance[pep_id] = [transcript_id]
                            else:
                                pep_to_transcript_accordance[pep_id].append(transcript_id)
                                print("Genbank record for %s contains several CDS features" % pep_id)
                                pep_with_several_CDS_features.append(pep_id)
                            if transcript_id in transcript_ids:
                                print("Repeated transcript id: %s" % transcript_id)
                                continue
                            transcript_ids.append(transcript_id)
                        except:
                            print("Transcript id for %s was not found" % pep_id)
                            pep_without_transcripts.append(pep_id)

        pep_with_several_CDS_features.write("%s.pep_with_several_CDS.ids" % output_prefix)
        pep_without_transcripts.write("%s.pep_without_transcripts.ids" % output_prefix)
        transcript_ids.write("%s.transcripts.ids" % output_prefix)

        mito_pep_with_several_CDS_features.write("%s.mito.pep_with_several_CDS.ids" % output_prefix)
        mito_pep_without_transcripts.write("%s.mito.pep_without_transcripts.ids" % output_prefix)
        mito_transcript_ids.write("%s.mito.transcripts.ids" % output_prefix)

        number_of_transcripts = len(transcript_ids) + len(mito_transcript_ids)
        print("%i transcripts were found" % number_of_transcripts)

        pep_to_transcript_accordance.write("%s.pep_to_transcript.accordance" % output_prefix, splited_values=True)
        mito_pep_to_transcript_accordance.write("%s.mito.pep_to_transcript.accordance" % output_prefix, splited_values=True)

        transcript_ranges = np.append(np.arange(0, number_of_transcripts, download_chunk_size), [number_of_transcripts])

        print("Downloading transcripts...")
        for i in range(0, len(transcript_ranges)-1):
            print("Downloading chunk %i" % i)
            transcript_tmp_file = "%s/chunk_%i" % (transcript_temp_dir, i)
            self.efetch("nuccore", transcript_ids[transcript_ranges[i]:transcript_ranges[i+1]],
                        transcript_tmp_file, rettype="gb", retmode="text",
                        log_file="%s.transcripts.efetch.log" % output_prefix)

        os.system("cat %s/* > %s" % (transcript_temp_dir, transcript_file))

        transcript_dict = SeqIO.index_db("%stmp_1.idx" % output_directory, transcript_file, format="genbank")

        print("Downloading mitochondrial transcripts...")
        self.efetch("nuccore", mito_transcript_ids, mito_transcript_file, rettype="gb", retmode="text",
                    log_file="%s.mito.transcripts.efetch.log" % output_prefix)
        mito_transcript_dict = SeqIO.index_db("%stmp_2.idx" % output_directory, mito_transcript_file, format="genbank")

        actual_protein_ids_from_transcripts = IdList()
        actual_mito_protein_ids_from_transcripts = IdList()
        actual_absent_protein_ids_from_transcripts = IdList()
        actual_absent_mito_protein_ids_from_transcripts = IdList()
        actual_pep_to_transcript_accordance = SynDict()
        actual_mito_pep_to_transcript_accordance = SynDict()
        actual_transcript_ids = IdList()
        actual_mito_transcript_ids = IdList()

        cds_records_dict = OrderedDict()
        for transcript_id in transcript_dict:
            CDS_number = 0
            for feature in transcript_dict[transcript_id].features:
                if feature.type == "CDS":
                    CDS_number += 1
            if CDS_number > 1:
                CDS_counter = 1
                for feature in transcript_dict[transcript_id].features:
                    if feature.type == "CDS":
                        feature_seq = feature.extract(transcript_dict[transcript_id].seq)
                        feature_id = "%s.%i" % (transcript_id, CDS_counter)
                        if "protein_id" in feature.qualifiers:
                            description = "protein=%s" % feature.qualifiers["protein_id"][0]
                            actual_protein_ids_from_transcripts.append(feature.qualifiers["protein_id"][0])
                            actual_pep_to_transcript_accordance[feature.qualifiers["protein_id"][0]] = [feature_id]
                            actual_transcript_ids.append(feature_id)
                        else:
                            print("Corresponding protein id was not found for %s " % transcript_id)

                        cds_records_dict[feature_id] = SeqRecord(seq=feature_seq, id=feature_id, description=description)
                        CDS_counter += 1
            else:
                for feature in transcript_dict[transcript_id].features:
                    if feature.type == "CDS":
                        feature_seq = feature.extract(transcript_dict[transcript_id].seq)
                        feature_id = transcript_id
                        if "protein_id" in feature.qualifiers:
                            description = "protein=%s" % feature.qualifiers["protein_id"][0]
                            actual_protein_ids_from_transcripts.append(feature.qualifiers["protein_id"][0])
                            actual_pep_to_transcript_accordance[feature.qualifiers["protein_id"][0]] = [feature_id]
                            actual_transcript_ids.append(feature_id)
                        else:
                            print("Corresponding protein id was not found for %s " % transcript_id)

                        cds_records_dict[feature_id] = SeqRecord(seq=feature_seq, id=feature_id, description=description)
                        break
        SeqIO.write(self.record_by_id_generator(cds_records_dict), "%s.all.cds" % output_prefix, format="fasta")
        SeqIO.write(self.record_by_id_generator(peptide_dict, id_list=actual_protein_ids_from_transcripts),
                    "%s.with_cds.pep" % output_prefix, format="fasta")
        SeqIO.write(self.record_by_id_generator(cds_records_dict, id_list=actual_transcript_ids),
                    "%s.with_pep.cds" % output_prefix, format="fasta")

        actual_pep_to_transcript_accordance.write("%s.pep_to_transcript.actual.accordance" % output_prefix, splited_values=True)

        mito_cds_records_dict = OrderedDict()
        for transcript_id in mito_transcript_dict:
            CDS_number = 0
            #print transcript_id
            for feature in mito_transcript_dict[transcript_id].features:
                if feature.type == "CDS":
                    CDS_number += 1
            #print CDS_number
            if CDS_number > 1:
                CDS_counter = 1
                for feature in mito_transcript_dict[transcript_id].features:
                    if feature.type == "CDS":
                        feature_seq = feature.extract(mito_transcript_dict[transcript_id].seq)
                        feature_id = "%s.%i" % (transcript_id, CDS_counter)
                        if "protein_id" in feature.qualifiers:
                            description = "protein=%s" % feature.qualifiers["protein_id"][0]
                            actual_mito_protein_ids_from_transcripts.append(feature.qualifiers["protein_id"][0])
                            actual_mito_pep_to_transcript_accordance[feature.qualifiers["protein_id"][0]] = [feature_id]
                            actual_mito_transcript_ids.append(feature_id)
                        else:
                            print("Corresponding protein id was not found for %s " % transcript_id)
                        CDS_counter += 1
                        mito_cds_records_dict[feature_id] = SeqRecord(seq=feature_seq, id=feature_id, description=description)

            else:
                for feature in mito_transcript_dict[transcript_id].features:
                    if feature.type == "CDS":
                        feature_seq = feature.extract(mito_transcript_dict[transcript_id].seq)
                        feature_id = transcript_id
                        if "protein_id" in feature.qualifiers:
                            description = "protein=%s" % feature.qualifiers["protein_id"][0]
                            actual_mito_protein_ids_from_transcripts.append(feature.qualifiers["protein_id"][0])
                            actual_mito_pep_to_transcript_accordance[feature.qualifiers["protein_id"][0]] = [feature_id]
                            actual_mito_transcript_ids.append(feature_id)
                        else:
                            print("Corresponding protein id was not found for %s " % transcript_id)

                        mito_cds_records_dict[feature_id] = SeqRecord(seq=feature_seq, id=feature_id, description=description)
                        break
        SeqIO.write(self.record_by_id_generator(mito_cds_records_dict), "%s.mito.all.cds" % output_prefix, format="fasta")
        SeqIO.write(self.record_by_id_generator(peptide_dict, id_list=actual_mito_protein_ids_from_transcripts),
                    "%s.mito.with_cds.pep" % output_prefix, format="fasta")
        SeqIO.write(self.record_by_id_generator(mito_cds_records_dict, id_list=actual_mito_transcript_ids),
                    "%s.mito.with_pep.cds" % output_prefix, format="fasta")

        #print actual_mito_pep_to_transcript_accordance
        actual_mito_pep_to_transcript_accordance.write("%s.mito.pep_to_transcript.actual.accordance" % output_prefix, splited_values=True)

        stat_string = "Input protein ids\t %i\n" % number_of_ids
        stat_string += "Downloaded proteins\t%i\n" % len(downloaded_protein_ids)
        stat_string += "Downloaded transcripts\t%i\n" % (len(transcript_dict) + len(mito_transcript_dict))
        stat_string += "Valid protein and CDS pair\t%i\n" % (len(actual_pep_to_transcript_accordance) +
                                                             len(actual_mito_pep_to_transcript_accordance))
        print(stat_string)

        with open("%s.stats" % output_prefix, "w") as stat_fd:
            stat_fd.write(stat_string)

        description_string = """DESCRIPTION OF FILES CREATED DURING DOWNLOADa AND EXTRACTION OF CDS FOR PROTEIN IDS FROM NCBI

* - prefix of files

RESULTING FILES
*.with_cds.pep                                 Final protein set for nuclear genes
*.with_pep.cds                                 Final CDS set for nuclear genes
*.pep_to_transcript.actual.accordance          Correspondence file between proteins and CDSs for nuclear genes

*.mito.with_cds.pep                            Final protein set for mitochondrial genes
*.mito.with_pep.cds                            Final CDS set for mitochondrial genes
*.mito.pep_to_transcript.actual.accordance     Correspondence file between proteins
                                               and CDSs for mitochondrial genes

INTERMIDIATE FILES AND DIRECTORIES

temp_proteins/                                 Directory with downloaded proteins
temp_transcripts/                              Directory with downloaded transcripts

*.pep.genbank                                  Merged file with downloaded nucleolar proteins

*.downloaded.ids                               Ids of downloaded proteins
*.not_downloaded.ids                           Ids of not downloaded proteins

*.trascript.genbank                            Merged file with downloaded nucleolar transcripts
*.trascript.mito.genbank                       Merged file with downloaded mitochondrial
                                               transcripts

*.mito.all.cds
*.mito.pep_to_transcript.accordance
*.mito.pep_without_transcripts.ids
*.mito.pep_with_several_CDS.ids
*.mito.transcripts.ids

*.all.cds
*.pep_to_transcript.accordance
*.pep_without_transcripts.ids
*.pep_with_several_CDS.ids

*.stats                                       Statistics
*.transcripts.ids
"""

        with open("%sDESCRIPTION" % output_directory, "w") as descr_fd:
            descr_fd.write(description_string)

        for filename in "tmp.idx", "tmp_1.idx", "tmp_2.idx":
            os.remove("%s%s" % (output_directory, filename))

    def get_cds_for_proteins_from_id_file(self, protein_id_file, output_prefix, download_chunk_size=100):
        pep_ids = IdList()
        pep_ids.read(protein_id_file)

        self.get_cds_for_proteins(pep_ids, output_prefix, download_chunk_size=download_chunk_size)

    @staticmethod
    def get_taxonomy(taxa_list, output_file, email, input_type="latin"):
        Entrez.email = email
        out_file = open(output_file, "w")
        out_file.write("#species\trank\tlatin_name\tlineage\n")

        species_syn_dict = SynDict()

        if input_type == "latin":
            for taxon in taxa_list:
                print("Handling %s" % taxon)
                summary = Entrez.read(Entrez.esearch(db="taxonomy", term=taxon))
                if summary:
                    id_list = summary["IdList"]
                    species_syn_dict[taxon] = []
                    for id in id_list:
                        print("\tHandling %s" % id)
                        record = Entrez.read(Entrez.efetch(db="taxonomy", id=id, retmode="xml"))
                        #print record
                        out_file.write("%s\t%s\t%s\t%s\n" % (taxon,
                                                             record[0]["Rank"],
                                                             record[0]['ScientificName'],
                                                             record[0]["Lineage"]))

                        species_syn_dict[taxon].append(record[0]['ScientificName'])
                        #species_set.add(record[0]["Species"])
        elif input_type == "id":
            for taxon in taxa_list:
                print("Handling %s" % taxon)
                species_syn_dict[taxon] = []
                #print taxon
                record = Entrez.read(Entrez.efetch(db="taxonomy", id=taxon, retmode="xml"))
                #print record
                out_file.write("%s\t%s\t%s\t%s\n" % (taxon,
                                                     record[0]["Rank"],
                                                     record[0]['ScientificName'],
                                                     record[0]["Lineage"]))
                #print record[0]
                species_syn_dict[taxon].append(record[0]['ScientificName'])
                #species_set.add(record[0]["Species"])
        out_file.close()
        return species_syn_dict

    @staticmethod
    def get_subtaxa_for_taxa(taxa_list, email, output_prefix, subtaxa_rank=('species'), verbose=False):
        Entrez.email = email
        taxonomy_file = "%s.taxonomy" % output_prefix
        species_syn_file = "%s.species.ids" % output_prefix

        subtaxa_rank_list = subtaxa_rank.split(",") if isinstance(subtaxa_rank, str) else subtaxa_rank
        species_syn_dict = SynDict()
        with open(taxonomy_file, "w") as taxa_fd:
            taxa_fd.write("#species\trank\tlatin_name\tcommon_name\tlineage\n")
            for taxon in taxa_list:
                print("Handling %s..." % taxon)
                summary = Entrez.read(Entrez.esearch(db="taxonomy",
                                                     term="%s[subtree] AND (%s)" % (taxon, "[rank] OR ".join(subtaxa_rank_list) if len(subtaxa_rank_list) > 4 else "%s[rank]" % subtaxa_rank_list[0]), retmax=4000))

                if verbose:
                    print (summary)

                if summary:
                    id_list = summary["IdList"]
                    for id in id_list:
                        print("handling %s..." % id)
                        record = Entrez.read(Entrez.efetch(db="taxonomy", id=id, retmode="xml"))
                        if verbose:
                            print(record)
                        common_name = "NA"
                        if 'OtherNames' in record[0]:
                            if "GenbankCommonName" in record[0]['OtherNames']:
                                common_name = record[0]['OtherNames']["GenbankCommonName"]

                        taxa_fd.write("%s\t%s\t%s\t%s\n" % (record[0]['ScientificName'],
                                                                record[0]["Rank"],
                                                                common_name,
                                                                record[0]["Lineage"]))

                        species_syn_dict[record[0]['ScientificName']] = common_name

        species_syn_dict.write(filename=species_syn_file, header="#latin_name\tcommon_name\n")
        return species_syn_dict

    def get_taxonomy_from_id_file(self, taxa_file, output_file, email, input_type="latin"):

        taxa_list = IdList(filename=taxa_file)

        return self.get_taxonomy(taxa_list, output_file, email, input_type=input_type)

    def get_protein_info_from_ncbi_gff(self, ncbi_gff, output_prefix):
        ncbi_gff_collection = CollectionGFF
        for record in ncbi_gff_collection.gff_simple_generator(ncbi_gff):
            pass

    @staticmethod
    def extract_gene_info_from_ncbi_gff(input_gff, output, feature_type_list=["gene"]):
        with open(input_gff, "r") as in_fd:
            with open(output, 'w') as out_fd:
                out_fd.write("#GeneID\tSource\tName\tGeneSynonym\tGeneBiotype\tDbxref\tDescription\n")
                for line in in_fd:
                    if line[0] == "#":
                        continue
                    tmp_list = line.strip().split("\t")
                    if tmp_list[2] not in feature_type_list:
                        continue
                    source = tmp_list[1]

                    annotation_dict = AnnotationsRoutines.parse_gff_annotation_string_to_dict(tmp_list[8])
                    if "Dbxref" in annotation_dict:
                        for entry in annotation_dict["Dbxref"]:
                            if "GeneID" in entry:
                                geneid = entry.split(":")[1]
                                break
                            else:
                                geneid = "."
                        dbxref = ",".join(annotation_dict["Dbxref"])
                    else:
                        geneid = "."
                        dbxref = "."
                    name = ",".join(annotation_dict["Name"] if "Name" in annotation_dict else ".")
                    description = ",".join(annotation_dict["description"] if "description" in annotation_dict else ".")
                    gene_biotype = ",".join(annotation_dict["gene_biotype"] if "gene_biotype" in annotation_dict else ".")
                    gene_synonym = ",".join(annotation_dict["gene_synonym"] if "gene_synonym" in annotation_dict else ".")
                    out_fd.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (geneid, source, name, gene_synonym,
                                                                   gene_biotype, dbxref, description))

    def get_taxa_genomes_summary(self, taxa, email, output_directory, output_prefix,
                                 max_ids_per_query=8000, max_download_attempts=500,
                                 min_scaffold_n50=None, min_contig_n50=None, max_scaffold_l50=None,
                                 max_contig_l50=None, max_contig_count=None, max_scaffold_count=None,
                                 max_chromosome_count=None, min_chromosome_count=None, max_unlocalized_scaffolds=None,
                                 max_unplaced_scaffolds=None, max_total_length=None, min_total_length=None,
                                 max_ungapped_length=None, min_ungapped_length=None,
                                 no_ambiguous_species=True):
        self.safe_mkdir(output_directory)
        Entrez.email = email
        taxa_list = taxa if isinstance(taxa, Iterable) else [taxa]

        all_files_dir = "%s%s/" % (self.check_path(output_directory), "all")
        nonambiguous_species_all_dir = "%snonambiguous_species_all/" % self.check_path(output_directory)
        ambiguous_species_all_dir = "%s%s/" % (self.check_path(output_directory), "ambiguous_species_all")
        chromosome_lvl_dir = "%s%s/" % (self.check_path(output_directory), "chromosome_lvl")
        non_chromosome_lvl_dir = "%s%s/" % (self.check_path(output_directory), "nonchromosome_lvl")

        filtered_by_integrity_dir = "%s%s/" % (self.check_path(output_directory), "passed_integrity_filters")
        filtered_out_by_integrity_dir = "%s%s/" % (self.check_path(output_directory), "not_passed_integrity_filters")

        stat_dir = "%s%s/" % (self.check_path(output_directory), "stat")
        taxa_stat_dir = "%s%s/" % (self.check_path(output_directory), "taxa_stat")
        for subdir in (all_files_dir, chromosome_lvl_dir, non_chromosome_lvl_dir, stat_dir,
                       taxa_stat_dir, nonambiguous_species_all_dir, ambiguous_species_all_dir):
            self.safe_mkdir(subdir)

        filter_by_integrity = min_scaffold_n50 or min_contig_n50 or max_scaffold_l50 or max_contig_l50 \
                              or max_contig_count or max_scaffold_count or max_chromosome_count \
                              or min_chromosome_count or max_unlocalized_scaffolds \
                              or max_unplaced_scaffolds or max_total_length or min_total_length \
                              or max_ungapped_length or min_ungapped_length

        if filter_by_integrity:
            for subdir in (filtered_by_integrity_dir, filtered_out_by_integrity_dir):
                self.safe_mkdir(subdir)

        for taxon in taxa_list:
            search_term = "%s[Orgn]" % taxon

            attempt_counter = 1
            while True:
                try:
                    summary = Entrez.read(Entrez.esearch(db="genome", term=search_term, retmax=10000, retmode="xml"))
                    break
                except URLError:
                    if attempt_counter > max_download_attempts:
                        URLError("Network problems. Maximum attempt number is exceeded")
                    print("URLError. Retrying... Attempt %i" % attempt_counter)
                    attempt_counter += 1

            print("Were found %s species" % summary["Count"])
            #print summary

            taxon_stat_file = "%s/%s.stat" % (taxa_stat_dir, taxon.replace(" ", "_"))
            taxon_stat_dict = TwoLvlDict()

            for species_id in summary["IdList"]: #[167] :
                print("Handling species id %s " % species_id)

                species_stat_file = "%s/%s.stat" % (stat_dir, species_id)
                species_stat_dict = TwoLvlDict()
                species_stat_dict[species_id] = OrderedDict()

                taxon_stat_dict[species_id] = OrderedDict()

                for stat in "all", "chromosome_lvl", "non_chromosome_lvl":
                    species_stat_dict[species_id][stat] = 0
                    taxon_stat_dict[species_id][stat] = 0
                #species_summary = Entrez.read(Entrez.esummary(db="genome", id=species_id, retmax=10000, retmode="xml"))
                #print species_summary

                # get assemblies linked with genome of species

                attempt_counter = 1
                while True:
                    try:
                        assembly_links = Entrez.read(Entrez.elink(dbfrom="genome", id=species_id, retmode="xml",
                                                                  retmax=10000, linkname="genome_assembly"))
                        break
                    except URLError:
                        if attempt_counter > max_download_attempts:
                            URLError("Network problems. Maximum attempt number is exceeded")
                        print("URLError. Retrying... Attempt %i" % attempt_counter)
                        attempt_counter += 1

                assembly_number = len(assembly_links)
                #print links
                #print links[0]["LinkSetDb"][0]["Link"]
                if assembly_links:
                    if "LinkSetDb" in assembly_links[0]:
                        if assembly_links[0]["LinkSetDb"]:
                            if "Link" in assembly_links[0]["LinkSetDb"][0]:
                                assembly_ids = [id_dict["Id"] for id_dict in assembly_links[0]["LinkSetDb"][0]["Link"]]
                            else:
                                continue
                        else:
                            continue
                    else:
                        continue
                else:
                    continue
                number_of_ids = len(assembly_ids)

                print("\tFound %i assemblies" % number_of_ids)

                id_group_edges = np.arange(0, number_of_ids+1, max_ids_per_query)

                if id_group_edges[-1] != number_of_ids:
                    id_group_edges = np.append(id_group_edges, number_of_ids)

                number_of_id_groups = len(id_group_edges) - 1

                #print len(assembly_links[0]["LinkSetDb"][0]["Link"])
                #print assembly_ids
                #print len(assembly_ids)
                #assembly_dict = TwoLvlDict()
                #assemblies_with_ambiguous_taxonomies = SynDict()
                #summaries = Entrez.read(Entrez.esummary(db="assembly", id=",".join(assembly_ids), retmode="xml"))

                summary_list = None
                for i in range(0, number_of_id_groups):
                    print("\tDownloading summary about assemblies %i - %i" % (id_group_edges[i]+1, id_group_edges[i+1]))
                    #print len(assembly_ids[id_group_edges[i]:id_group_edges[i+1]])
                    summaries = Entrez.read(Entrez.esummary(db="assembly",
                                                            id=",".join(assembly_ids[id_group_edges[i]:id_group_edges[i+1]]),
                                                            retmode="xml"), validate=False)
                    tmp_summary_list = AssemblySummaryList(entrez_summary_biopython=summaries)
                    summary_list = (summary_list + tmp_summary_list) if summary_list else tmp_summary_list

                print("\tDownloaded %i" % len(summary_list))

                if len(summary_list) != number_of_ids:
                    print("\tWARNING:Not all assemblies were downloaded")
                    """
                    print("\tFollowing assemblies were not downloaded(ids):%s" % ",".join(set()))
                    """

                if summary_list:
                    species_stat_dict[species_id]["all"] = len(summary_list)
                    taxon_stat_dict[species_id]["all"] = len(summary_list)
                    output_file = "%s%s.genome.summary" % ((output_prefix + ".") if output_prefix else "", species_id)
                                                           #summary_list[0]['SpeciesName'].replace(" ", "_"))

                    all_output_file = "%s/%s" % (all_files_dir, output_file)
                    chromosome_lvl_output_file = "%s/%s" % (chromosome_lvl_dir, output_file)
                    non_chromosome_lvl_output_file = "%s/%s" % (non_chromosome_lvl_dir, output_file)
                    nonambiguous_species_output_file = "%s/%s" % (nonambiguous_species_all_dir, output_file)
                    ambiguous_species_output_file = "%s/%s" % (ambiguous_species_all_dir, output_file)
                    chromosome_lvl_summary_list, non_chromosome_lvl_summary_list = summary_list.filter_non_chrom_level_genomes()
                    filtered_by_integrity_file = "%s/%s" % (filtered_by_integrity_dir, output_file)
                    filtered_out_by_integrity_file = "%s/%s" % (filtered_out_by_integrity_dir, output_file)

                    species_stat_dict[species_id]["chromosome_lvl"] = len(chromosome_lvl_summary_list)
                    taxon_stat_dict[species_id]["chromosome_lvl"] = len(chromosome_lvl_summary_list)
                    species_stat_dict[species_id]["non_chromosome_lvl"] = len(non_chromosome_lvl_summary_list)
                    taxon_stat_dict[species_id]["non_chromosome_lvl"] = len(non_chromosome_lvl_summary_list)

                    print("\tChromosome level assemblies %i" % species_stat_dict[species_id]["chromosome_lvl"])
                    print("\tNon chromosome level assemblies %i" % species_stat_dict[species_id]["non_chromosome_lvl"])

                    if chromosome_lvl_summary_list:
                        chromosome_lvl_summary_list.write(chromosome_lvl_output_file)

                    if non_chromosome_lvl_summary_list:
                        non_chromosome_lvl_summary_list.write(non_chromosome_lvl_output_file)

                    nonambiguous_species_summary_list, ambiguous_species_summary_list = summary_list.filter_ambiguous_species()
                    #print(len(nonambiguous_species_summary_list), len(ambiguous_species_summary_list))
                    species_stat_dict[species_id]["nonambiguous_species"] = len(nonambiguous_species_summary_list)
                    species_stat_dict[species_id]["ambiguous_species"] = len(ambiguous_species_summary_list)
                    print("\tAmbiguous species %i" % species_stat_dict[species_id]["ambiguous_species"])
                    if nonambiguous_species_summary_list:
                        nonambiguous_species_summary_list.write(nonambiguous_species_output_file)
                    if ambiguous_species_summary_list:
                        ambiguous_species_summary_list.write(ambiguous_species_output_file)

                    summary_list.write(all_output_file)

                    if filter_by_integrity:
                        filtered_by_integrity, filtered_out_by_integrity = summary_list.filter_by_integrity(min_scaffold_n50=min_scaffold_n50,
                                                                                                            min_contig_n50=min_contig_n50,
                                                                                                            max_scaffold_l50=max_scaffold_l50,
                                                                                                            max_contig_l50=max_contig_l50,
                                                                                                            max_contig_count=max_contig_count,
                                                                                                            max_scaffold_count=max_scaffold_count,
                                                                                                            max_chromosome_count=max_chromosome_count,
                                                                                                            min_chromosome_count=min_chromosome_count,
                                                                                                            max_unlocalized_scaffolds=max_unlocalized_scaffolds,
                                                                                                            max_unplaced_scaffolds=max_unplaced_scaffolds,
                                                                                                            max_total_length=max_total_length,
                                                                                                            min_total_length=min_total_length,
                                                                                                            max_ungapped_length=max_ungapped_length,
                                                                                                            min_ungapped_length=min_ungapped_length,
                                                                                                            no_ambiguous_species=no_ambiguous_species)
                        species_stat_dict[species_id]["filtered_by_integrity"] = len(filtered_by_integrity)
                        species_stat_dict[species_id]["filtered_out_by_integrity"] = len(filtered_out_by_integrity)
                        if filtered_by_integrity:
                            filtered_by_integrity.write(filtered_by_integrity_file)
                        if filtered_out_by_integrity:
                            filtered_out_by_integrity.write(filtered_out_by_integrity_file)
                        print("\tPassed integrity filters %i" % species_stat_dict[species_id]["filtered_by_integrity"])
                species_stat_dict.write(species_stat_file)

                print("\n\n")

            taxon_stat_dict.write(taxon_stat_file)

            """
                for tmp in summary_set:
                    #print "aaaaaaaaaaa"

                    tmp = Entrez.read(Entrez.esummary(db="assembly",
                                                                                   id=assembly_id,
                                                                            retmode="xml"))

                    #print tmp
                    #assembly_summary = AssemblySummary(tmp)
                    #print assembly_summary

                    #print assembly_summary.SpeciesName


                    taxonomy_links = Entrez.read(Entrez.elink(dbfrom="assembly", id=assembly_id, retmode="xml",
                                                              retmax=10000, linkname="assembly_taxonomy"))
                    #print taxonomy_links
                    taxonomy_ids = [id_dict["Id"] for id_dict in taxonomy_links[0]["LinkSetDb"][0]["Link"]]

                    if len(taxonomy_ids) > 1:
                        print( "WARNING: more than one taxonomy id for assembly %s" % assembly_ids)
                        assemblies_with_ambiguous_taxonomies[assembly_id] = taxonomy_ids
                    #print taxonomy_ids

                    taxonomy_record = Entrez.read(Entrez.efetch(db="taxonomy", id=taxonomy_ids[0], retmode="xml"))

                    #print taxonomy_record
                    scientific_name = taxonomy_record[0]["ScientificName"]
                    lineage = taxonomy_record[0]["Lineage"]
                    #common_name_list = taxonomy_record[0]["CommonName"]
                    print (taxonomy_record[0]["ScientificName"])
                """


"""
assembly_nuccore 	Nucleotide 	Nucleotide 	Nucleotide 	5000
assembly_nuccore_insdc 	Nucleotide INSDC 	Nucleotide INSDC 	Nucleotide INSDC (GenBank) 	5000
assembly_nuccore_refseq 	Nucleotide RefSeq 	Nucleotide RefSeq 	Nucleotide RefSeq 	5000
assembly_nuccore_wgsmaster 	WGS Master 	WGS Master 	WGS Master 	5000
assembly_nuccore_wgscontig 	WGS contigs 	WGS contigs 	WGS contigs 	5000
assembly_nuccore_insdc 	Nucleotide INSDC 	Nucleotide INSDC 	Nucleotide INSDC (GenBank) 	5000
assembly_nuccore_refseq 	Nucleotide RefSeq 	Nucleotide RefSeq 	Nucleotide RefSeq 	5000
assembly_nuccore_wgsmaster 	WGS Master 	WGS Master 	WGS Master
"""

"""

#Example of assembly summary

{u'DocumentSummarySet':
     DictElement({u'DbBuild': 'Build161130-1600.1',
                  u'DocumentSummary': [DictElement({u'ChainId': '1886755',
                                                    u'AsmUpdateDate': '2016/11/28 00:00',
                                                    u'GbUid': '3745278',
                                                    u'PropertyList': ['full-genome-representation', 'has-chromosome', 'has-plasmid', 'latest', 'latest_genbank'],
                                                    u'SubmissionDate': '2016/11/28 00:00',
                                                    u'GB_BioProjects': [{u'BioprojectAccn': 'PRJNA353939',
                                                                         u'BioprojectId': '353939'}],
                                                    u'AssemblyAccession': 'GCA_001886755.1',
                                                    u'LastMajorReleaseAccession': 'GCA_001886755.1',
                                                    u'Synonym': {u'RefSeq': '',
                                                                 u'Genbank': 'GCA_001886755.1',
                                                                 u'Similarity': ''},
                                                    u'FtpPath_RefSeq': '',
                                                    u'AsmReleaseDate_RefSeq': '1/01/01 00:00',
                                                    u'AssemblyType': 'haploid',
                                                    u'AssemblyDescription': '',
                                                    u'AsmReleaseDate_GenBank': '2016/11/28 00:00',
                                                    u'RS_Projects': [],
                                                    u'RefSeq_category': 'na',
                                                    u'PartialGenomeRepresentation': 'false',
                                                    u'Coverage': '0',
                                                    u'AssemblyClass': 'haploid',
                                                    u'ExclFromRefSeq': [],
                                                    u'SeqReleaseDate': '2016/11/28 00:00',
                                                    u'AnomalousList': [],
                                                    u'AssemblyStatus': 'Complete Genome',
                                                    u'AsmReleaseDate': '2016/11/28 00:00',
                                                    u'ReleaseLevel': 'Major',
                                                    u'Taxid': '562',
                                                    u'LastUpdateDate': '2016/11/28 00:00',
                                                    u'RsUid': '',
                                                    u'FromType': '',
                                                    u'WGS': '',
                                                    u'GB_Projects': ['353939'],
                                                    u'BioSampleId': '6029894',
                                                    u'AssemblyName': 'ASM188675v1',
                                                    u'EnsemblName': '',
                                                    u'Organism': 'Escherichia coli (E. coli)',
                                                    u'Biosource': {u'Isolate': '',
                                                                   u'InfraspeciesList': [{u'Sub_type': 'strain',
                                                                                          u'Sub_value': 'MRSN346595'}],
                                                                   u'Sex': ''},
                                                    u'SpeciesName': 'Escherichia coli',
                                                    u'BioSampleAccn': 'SAMN06029894',
                                                    u'FtpPath_GenBank': 'ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCA/001/886/755/GCA_001886755.1_ASM188675v1',
                                                    u'SpeciesTaxid': '562',
                                                    u'UCSCName': '',
                                                    u'Primary': '3745268',
                                                    u'Meta': ' <Stats> <Stat category="alt_loci_count" sequence_tag="all">0</Stat> <Stat category="chromosome_count" sequence_tag="all">1</Stat> <Stat category="contig_count" sequence_tag="all">6</Stat> <Stat category="contig_l50" sequence_tag="all">1</Stat> <Stat category="contig_n50" sequence_tag="all">4796421</Stat> <Stat category="non_chromosome_replicon_count" sequence_tag="all">5</Stat> <Stat category="replicon_count" sequence_tag="all">6</Stat> <Stat category="scaffold_count" sequence_tag="all">6</Stat> <Stat category="scaffold_count" sequence_tag="placed">6</Stat> <Stat category="scaffold_count" sequence_tag="unlocalized">0</Stat> <Stat category="scaffold_count" sequence_tag="unplaced">0</Stat> <Stat category="scaffold_l50" sequence_tag="all">1</Stat> <Stat category="scaffold_n50" sequence_tag="all">4796421</Stat> <Stat category="total_length" sequence_tag="all">5116627</Stat> <Stat category="ungapped_length" sequence_tag="all">5116627</Stat> </Stats> <FtpSites>   <FtpPath type="GenBank">ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCA/001/886/755/GCA_001886755.1_ASM188675v1</FtpPath> </FtpSites> <assembly-level>5</assembly-level> <assembly-status>Complete Genome</assembly-status> <representative-status>na</representative-status> <submitter-organization>Walter Reed Army Institute of Research</submitter-organization>    ',
                                                    u'NCBIReleaseDate': '2016/11/28 00:00',
                                                    u'SubmitterOrganization': 'Walter Reed Army Institute of Research',
                                                    u'RS_BioProjects': [],
                                                    u'SortOrder': '5C10018867559899'},
                                                    attributes={u'uid': u'894571'}
                                                   )
                                       ]
                  },
                 attributes={u'status': u'OK'}
                 )
 }
"""



#!/usr/bin/env python
__author__ = 'mahajrod'
from collections import OrderedDict
from RouToolPa.Collections.General import TwoLvlDict

# adapter for 2 versions of Coockiecutter

class CoockiecutterReport(OrderedDict):
    def __init__(self, coockiecutter_report_file, input_is_se=False):
        #self = OrderedDict()
        OrderedDict.__init__(self)
        self["left"] = OrderedDict()
        self["right"] = OrderedDict()
        self["se"] = OrderedDict()

        with open(coockiecutter_report_file, "r") as in_fd:
            tmp = ""

            for read_position in ("se", ) if input_is_se else ("left", "right"):
                while "ok" not in tmp:
                    tmp = in_fd.readline()

                while "\t" in tmp:
                    tmp = tmp.strip().split("\t")
                    if "%" in tmp[1]:
                        self[read_position][tmp[0]] = float(tmp[1].replace("%", ""))
                    else:
                        self[read_position][tmp[0]] = int(tmp[1])
                    tmp = in_fd.readline()

        self.match_key = "match" if (("match" in self["left"]) or ("match" in self["se"]) ) else "adapter"
        self.retained_pairs_key = "ok" if input_is_se else "pe" if "pe" in self["left"] else "paired-end reads"

        pos = "se" if input_is_se else "left"

        self.input_pairs = self[pos]["ok"] + self[pos][self.match_key] + (self[pos]["n"] if "n" in self[pos] else 0)
        self.retained_pairs = self[pos][self.retained_pairs_key]
        self.retained_pairs_fraction = float(self.retained_pairs) / float(self.input_pairs)


class CoockiecutterReportCollection(OrderedDict):
    def __init__(self, report_dict=None, file_dict=None):
        OrderedDict.__init__(self)
        if report_dict:
            for report_id in report_dict:
                self[report_id] = report_dict
        if file_dict:
            for file_id in file_dict:
                if file_id in self:
                    raise KeyError("Report with same id(%s) is already present in CoockiecutterReportCollection" % str(file_id))
                self[file_id] = CoockiecutterReport(file_dict[file_id])

    def get_general_stats(self):
        stat_dict = TwoLvlDict()

        for report_id in self:
            stat_dict[report_id] = OrderedDict()

            stat_dict[report_id]["input_pairs"] = self[report_id].input_pairs
            stat_dict[report_id]["pairs_without_adapters"] = self[report_id].retained_pairs
            stat_dict[report_id]["pairs_without_adapters_fraction"] = self[report_id].retained_pairs_fraction

        return stat_dict

    def write_general_stats(self, output_file):
        stat_dict = self.get_general_stats()
        stat_dict.write(output_file, absent_symbol=".", sort=False)



#!/usr/bin/env python
__author__ = 'Sergei F. Kliver'

import os
from RouToolPa.Tools.Abstract import Tool


class Sealer(Tool):
    def __init__(self, path="", max_threads=4):
        Tool.__init__(self, "abyss-sealer", path=path, max_threads=max_threads)

    def close_gaps(self, scaffold_fasta, forward_reads, reverse_reads, output_prefix,
                   kmer_size_list=(90, 80, 70, 60, 50, 40, 30), print_flanks=False, flank_len=None,
                   flank_distance=None, max_gap_len=None, bloom_size=None,
                   max_branches=None, dot_outfile=None, fix_errors=None, input_bloom_filter=None,
                   soft_mask_new_and_changed_bases=None, dont_soft_mask_bases=None, flank_mismatches=None,
                   max_mismatches=None, disable_all_limits=None, max_paths=None, trim_masked_from_ends=None,
                   do_not_discard_unchaste_reads=None, trim_quality=None):

        options = " -j %i" % self.threads
        options += " -S %s" %scaffold_fasta
        options += " -o %s" % output_prefix
        options += " %s" % (" ".join(map(lambda kmer_size: " -k%i" % kmer_size, kmer_size_list)))
        options += " --print-flanks" if print_flanks else ""
        options += " -L %i" % flank_len  if flank_len else ""
        options += " -D %i" % flank_distance if flank_distance else ""
        options += " -G %i" % max_gap_len if max_gap_len else ""
        options += " -b %s" % (bloom_size if isinstance(bloom_size, str) else "%i" % bloom_size) if bloom_size else ""
        options += " -B %i" % max_branches if max_branches else ""
        options += " -d %s" % dot_outfile if dot_outfile else ""
        options += " --fix-errors" if fix_errors else ""
        options += " -i %s" % input_bloom_filter if input_bloom_filter else ""
        options += " --mask" if soft_mask_new_and_changed_bases else ""
        options += " --no_mask" if dont_soft_mask_bases else ""
        options += " -m %i" % flank_mismatches if flank_mismatches else ""
        options += " -M %i" % max_mismatches if max_mismatches else ""
        options += " -P %i" % max_paths if max_paths else ""
        options += " -n" if disable_all_limits else ""
        options += " --trim-masked" if trim_masked_from_ends else ""
        options += " --no-chastity" if do_not_discard_unchaste_reads else ""
        options += " --trim-quality %i" % trim_quality if trim_quality else ""
        options += " %s" % forward_reads
        options += " %s" % reverse_reads

        self.execute(options)
        """
        Not implemented options

        --standard-quality       zero quality is `!' (33)
                               default for FASTQ and SAM files
        --illumina-quality       zero quality is `@' (64)
                               default for qseq and export files
        -r, --read-name=STR          only process reads with names that contain STR
        -s, --search-mem=N           mem limit for graph searches; multiply by the
                               number of threads (-j) to get the total mem used
                               for graph traversal [500M]
        -t, --trace-file=FILE        write graph search stats to FILE

        """

if __name__ == "__main__":
    pass



#!/usr/bin/env python

from RouToolPa.Tools.Abstract import Tool


class RAxML(Tool):
    def __init__(self, path="", max_threads=4):
        Tool.__init__(self, "raxml", path=path, max_threads=max_threads)

    def construct_tree_with_bootstrap(self,
                                      alignment_file,
                                      partition_file,
                                      output_prefix,
                                      model="GTRGAMMAI",
                                      bootstrap_replicates=100,
                                      outgroup=None,
                                      bootstrap_random_seed=12345,
                                      parsimony_random_seed=12345):
        options = " -f a"
        options += " -T %i" % self.threads
        options += " -x %s" % alignment_file
        options += " -n %s" % output_prefix
        options += " -o %s" % (outgroup if isinstance(outgroup, str) else ",".join(outgroup)) if outgroup else ""
        options += " -q %s" % partition_file
        options += " -# %s" % bootstrap_replicates
        options += " -x %i" % bootstrap_random_seed
        options += " -p %i" % parsimony_random_seed
        options += " -m %s" % model

        self.execute(options)



__author__ = 'mahajrod'
import os
import math
from collections import Iterable, OrderedDict
from math import sqrt

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
plt.ioff()
import numpy as np

from RouToolPa.Collections.General import TwoLvlDict, IdList, SynDict
from RouToolPa.Parsers.Abstract import Record, Collection
from RouToolPa.Routines import FileRoutines, MatplotlibRoutines
from Bio import AlignIO


class MultipleAlignmentStatRecord(Record):

    def __init__(self, aln_id, alignment, description=None, flags=None, gap_symbol="-", verbose=False):
        self.id = aln_id
        self.alignment = alignment
        self.description = description
        self.flags = flags
        self.gap_symbol = gap_symbol
        self.number_of_sequences = len(alignment)
        self.length = alignment.get_alignment_length()

        self.position_presence_matrix = self.get_position_presence_matrix(verbose=verbose)
        self.gap_counts, self.seq_lengths = self.count_gaps()
        self.N_counts = self.count_Ns()

        self.unique_pos_matrix_count, self.unique_pos_matrix_count_percent = self.count_unique_positions_per_sequence()

        #self.unique_insertion_counts, self.unique_gap_counts, self.unique_leading_counts, self.unique_trailing_counts = self.count_unique_positions_per_sequence()

    def __str__(self):
        """
        General stat string

        """
        """
        return "%s\t%i\t%i\t%i\t%i\t%i\t%f\t%i\t%f\t%i\t%f\t%i\t%f\t%i\t%f\n" % (self.id, self.number_of_sequences, self.length,
                                                                         min(self.seq_lengths.values()),
                                                                         max(self.seq_lengths.values()),
                                                                         max(self.unique_pos_matrix_count[:, 0]),
                                                                         max(self.unique_pos_matrix_count_percent[:, 0]),
                                                                         max(self.unique_pos_matrix_count[:, 1]),
                                                                         max(self.unique_pos_matrix_count_percent[:, 1]),
                                                                         max(self.unique_pos_matrix_count[:, 2]) if math.fabs(max(self.unique_pos_matrix_count[:, 2])) > math.fabs(min(self.unique_pos_matrix_count[:, 2])) else min(self.unique_pos_matrix_count[:, 2]),
                                                                         max(self.unique_pos_matrix_count_percent[:, 2]) if math.fabs(max(self.unique_pos_matrix_count_percent[:, 2])) > math.fabs(min(self.unique_pos_matrix_count_percent[:, 2])) else min(self.unique_pos_matrix_count_percent[:, 2]),
                                                                         max(self.unique_pos_matrix_count[:, 3]) if math.fabs(max(self.unique_pos_matrix_count[:, 3])) > math.fabs(min(self.unique_pos_matrix_count[:, 3])) else min(self.unique_pos_matrix_count[:, 3]),
                                                                         max(self.unique_pos_matrix_count_percent[:, 3]) if math.fabs(max(self.unique_pos_matrix_count_percent[:, 3])) > math.fabs(min(self.unique_pos_matrix_count_percent[:, 3])) else min(self.unique_pos_matrix_count_percent[:, 3]),
                                                                         max(self.unique_pos_matrix_count[:, 4]),
                                                                         max(self.unique_pos_matrix_count_percent[:, 4]))
        """
        return "%s\t%s\n" % (self.id, "\t".join(map(str, self.get_general_stats())))

    def count_gaps(self):

        gaps_dict = SynDict()
        seq_length_dict = SynDict()

        for row in range(0, self.number_of_sequences):
            sequence_id = self.alignment[row].id
            gaps_dict[sequence_id] = str(self.alignment[row].seq).count(self.gap_symbol)
            seq_length_dict[sequence_id] = self.length - gaps_dict[sequence_id]

        return gaps_dict, seq_length_dict

    def get_position_presence_matrix(self, verbose=True):

        number_of_sequences = len(self.alignment)
        alignment_length = len(self.alignment[0])
        # converting alignment to numpy letter array stored by columns!
        alignment_array = np.array([list(rec) for rec in self.alignment], np.character, order="F")

        if verbose:
            print("%i sequences in alignment" % number_of_sequences)
            print("%i columns in alignment" % alignment_length)

        position_presence_array = np.array([[0 for letter in rec.seq] for rec in self.alignment], int, order="F")

        for column in range(0, alignment_length):
            column_list = alignment_array[:, column]
            number_of_gaps = 0
            for element in column_list:
                if element == self.gap_symbol:
                    number_of_gaps += 1
            for row in range(0, number_of_sequences):
                position_presence_array[row, column] = -number_of_gaps if alignment_array[row, column] == self.gap_symbol else number_of_sequences - number_of_gaps

        return position_presence_array

    def count_Ns(self):
        Ns_dict = SynDict()
        for row in range(0, self.number_of_sequences):
            sequence_id = self.alignment[row].id
            Ns_dict[sequence_id] = self.alignment[row].seq.count("N")
        return Ns_dict

    def get_general_stats(self):
        return [self.number_of_sequences, self.length,
                min(self.seq_lengths.values()),
                max(self.seq_lengths.values()),
                max(self.unique_pos_matrix_count[:, 0]),
                max(self.unique_pos_matrix_count_percent[:, 0]),
                max(self.unique_pos_matrix_count[:, 1]),
                max(self.unique_pos_matrix_count_percent[:, 1]),
                max(self.unique_pos_matrix_count[:, 2]) if math.fabs(max(self.unique_pos_matrix_count[:, 2])) > math.fabs(min(self.unique_pos_matrix_count[:, 2])) else min(self.unique_pos_matrix_count[:, 2]),
                max(self.unique_pos_matrix_count_percent[:, 2]) if math.fabs(max(self.unique_pos_matrix_count_percent[:, 2])) > math.fabs(min(self.unique_pos_matrix_count_percent[:, 2])) else min(self.unique_pos_matrix_count_percent[:, 2]),
                max(self.unique_pos_matrix_count[:, 3]) if math.fabs(max(self.unique_pos_matrix_count[:, 3])) > math.fabs(min(self.unique_pos_matrix_count[:, 3])) else min(self.unique_pos_matrix_count[:, 3]),
                max(self.unique_pos_matrix_count_percent[:, 3]) if math.fabs(max(self.unique_pos_matrix_count_percent[:, 3])) > math.fabs(min(self.unique_pos_matrix_count_percent[:, 3])) else min(self.unique_pos_matrix_count_percent[:, 3]),
                max(self.unique_pos_matrix_count[:, 4]),
                max(self.unique_pos_matrix_count_percent[:, 4])]

    def count_unique_positions_per_sequence(self):
        unique_matrix_count = []
        unique_matrix_count_percent = []
        #unique_insertion_count_dict = SynDict()
        #unique_gap_count_dict = SynDict()

        #unique_leading_count_dict = SynDict()
        #unique_trailing_count_dict = SynDict()

        for row in range(0, self.number_of_sequences):
            sequence_id = self.alignment[row].id
            unique_gaps = 0
            unique_insertions = 0
            leading_unique_positions = 0
            trailing_unique_positions = 0
            unique_positions = 0
            for column in range(0, self.length):
                if self.position_presence_matrix[row, column] == 1:
                    unique_insertions += 1
                    unique_positions += 1
                elif self.position_presence_matrix[row, column] == -1:
                    unique_gaps += 1
                    unique_positions += 1

            if (self.position_presence_matrix[row, 0] == 1) or (self.position_presence_matrix[row, 0] == -1):
                for column in range(0, self.length):
                    if self.position_presence_matrix[row, 0] == self.position_presence_matrix[row, column]:
                        leading_unique_positions += self.position_presence_matrix[row, column]
                    else:
                        break

            if (self.position_presence_matrix[row, -1] == 1) or (self.position_presence_matrix[row, -1] == -1):

                for column in range(-1, -self.length, -1):
                    if self.position_presence_matrix[row, -1] == self.position_presence_matrix[row, column]:
                        trailing_unique_positions += self.position_presence_matrix[row, column]
                    else:
                        break

            #unique_insertion_count_dict[sequence_id] = unique_insertions
            #unique_gap_count_dict[sequence_id] = unique_gaps
            #unique_leading_count_dict[sequence_id] = leading_unique_positions
            #unique_trailing_count_dict[sequence_id] = trailing_unique_positions
            #print unique_trailing_count_dict[sequence_id]

            unique_matrix_count.append([unique_insertions,
                                        unique_gaps,
                                        leading_unique_positions,
                                        trailing_unique_positions,
                                        unique_positions])

            unique_matrix_count_percent.append([float(unique_insertions)/float(self.seq_lengths[sequence_id]),
                                                float(unique_gaps)/float(self.seq_lengths[sequence_id]),
                                                float(leading_unique_positions)/float(self.seq_lengths[sequence_id]),
                                                float(trailing_unique_positions)/float(self.seq_lengths[sequence_id]),
                                                float(unique_positions)/float(self.seq_lengths[sequence_id])])

        unique_matrix_count = np.array(unique_matrix_count, dtype=int)
        unique_matrix_count_percent = np.array(unique_matrix_count_percent, dtype=float)

        return unique_matrix_count, unique_matrix_count_percent

        #return unique_insertion_count_dict, unique_gap_count_dict, unique_leading_count_dict, unique_trailing_count_dict


class Metadata():
    def __init__(self, metadata=[]):
        self.metadata = metadata

    def add_metadata(self):
        pass

    def __str__(self):
        pass


class Header():
    def __init__(self):
        pass

    def __str__(self):
        pass


class MultipleAlignmentStatCollection(Collection):
    # TODO: develop this class to minimize new code in various collections
    def __init__(self, metadata=None, record_list=None, header=None, input_file=None, filetype="fasta", verbose=False,
                 general_stats_input=False):
        # metadata should be Metadata class
        # header should be Header class
        # record_list should be list of Record class
        self.general_stats_table = None
        self.general_stats_header = "#aln_id\tseq_number\taln_length\t" \
                                    "min_seq_len\tmax_seq_len\t" \
                                    "max_uniq_insertions\tmax_uniq_insertions,fraction\t" \
                                    "max_unique_gaps\tmax_unique_gaps,fraction\t" \
                                    "max_leading_unique_pos\tmax_leading_unique_pos,fraction\t" \
                                    "max_trailing_unique_pos\tmax_trailing_unique_pos,fraction\t" \
                                    "max_unique_pos\tmax_unique_pos,fraction\n"
        if not(input_file is None):
            self.records = OrderedDict()
            self.record_id_list = []
            if general_stats_input:
                self.read_general_stats_file(input_file)

            else:
                self.read(input_file, filetype=filetype, verbose=verbose)
                self.general_stats_table = self.get_general_stats()
        else:
            self.metadata = metadata
            self.records = record_list
            self.header = header

    def read(self, input_file, filetype="fasta", verbose=False):
        list_of_files = FileRoutines.make_list_of_path_to_files(input_file)
        for filename in list_of_files:
            if verbose:
                print("Parsing %s ..." % filename)
            directory, basename, extension = FileRoutines.split_filename(filename)
            try:
                self.records[basename] = MultipleAlignmentStatRecord(basename, alignment=AlignIO.read(filename, filetype))
                self.record_id_list.append(basename)
            except:
                raise ValueError("ERROR: Issues while parsing or calculating stats for %s!!!" % filename)
        # collectiontype-dependent function
        pass

    def read_general_stats_file(self, general_stats_file):
        self.record_id_list = []
        general_stats_list = []
        with open(general_stats_file, "r") as in_fd:
            in_fd.readline()
            for line in in_fd:
                line_list = line.strip("\n").split("\t")
                self.record_id_list.append(line_list[0])
                for i in [1, 2, 3, 4, 5, 7, 9, 11, 13]:
                    line_list[i] = int(line_list[i])
                for i in [6, 8, 10, 12, 14]:
                    line_list[i] = float(line_list[i])
                general_stats_list.append(line_list[1:])
        self.general_stats_table = np.array(general_stats_list)

    def get_general_stats(self):
        general_stats_list = []
        for record_id in self.records:
            #print self.records
            general_stats_list.append(self.records[record_id].get_general_stats())
        return np.array(general_stats_list)

    def write_general_stats(self, output):
        with open(output, "w") as out_fd:
            out_fd.write(self.general_stats_header)
            for record_id in self.records:
                #print self.records
                out_fd.write(str(self.records[record_id]))

    def write_stats(self, output_prefix):
        Ns_dict = TwoLvlDict()
        gaps_dict = TwoLvlDict()
        for record_id in self.records:
            Ns_dict[self.records[record_id].id] = self.records[record_id].N_counts
            gaps_dict[self.records[record_id].id] = self.records[record_id].gap_counts

        Ns_dict.write(out_filename="%s.N_counts" % output_prefix)
        gaps_dict.write(out_filename="%s.gaps_counts" % output_prefix)

    def draw_general_stats_distributions(self, output_prefix, figsize=(12, 6), extensions=("png", "svg"), dpi=300,
                                         logscale_heatmaps=True):

        nrows = 2
        ncols = 4

        #figure = plt.figure(figsize=(12, 6), dpi=dpi)
        #ax_array = figure.subplots(nrows=nrows, ncols=ncols, squeeze=False)
        percent_histogram_bin_number = 20

        min_seq_number_in_alignment = min(self.general_stats_table[:, 0])
        max_seq_number_in_alignment = max(self.general_stats_table[:, 0])
        seq_bin_number = max_seq_number_in_alignment - min_seq_number_in_alignment

        figure, ax_array = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, squeeze=False, dpi=dpi)

        # Histogram: distribution of sequence number in alignment
        print("Histogram: distribution of sequence number in alignment")

        MatplotlibRoutines.draw_histogram(self.general_stats_table[:, 0],
                                          output_prefix="%s.seq_number_distibution" % output_prefix,
                                          number_of_bins=None, width_of_bins=1,
                                          max_threshold=None, min_threshold=None, xlabel=None, ylabel="N of alignments",
                                          title="Distribution of\nsequence numbers",
                                          extensions=("png",), ylogbase=None, subplot=ax_array[0, 0], suptitle=None,
                                          close_figure=False, save_histovalues_only=True)
        # Histogram(logscaled): distribution of sequence number in alignment

        print("Histogram(logscaled): distribution of sequence number in alignment")
        MatplotlibRoutines.draw_histogram(self.general_stats_table[:, 0], output_prefix=None,
                                          number_of_bins=None, width_of_bins=1,
                                          max_threshold=None, min_threshold=None,
                                          xlabel="N of sequences\nin alignment", ylabel="N of alignments",
                                          title=None,
                                          extensions=("png",), ylogbase=10, subplot=ax_array[1, 0], suptitle=None,
                                          close_figure=False)

        #print self.general_stats_table[:, 4].astype(float) / self.general_stats_table[:, 2].astype(float) * 100

        # Heatmap: x: max_seq_len/aln_len, y: min_seq_len/aln_len
        print("Heatmap: x: max_seq_len/aln_len, y: min_seq_len/aln_len")
        MatplotlibRoutines.draw_percent_heatmap(self.general_stats_table[:, 3].astype(float) / self.general_stats_table[:, 1].astype(float) * 100,
                                                self.general_stats_table[:, 2].astype(float) / self.general_stats_table[:, 1].astype(float) * 100,
                                                output_prefix="%s.min_max_seq_len" % output_prefix, xlabel="Max seq len, % of aln",
                                                ylabel="Min seq len, % of aln", title=None,
                                                figsize=(8, 8), minimum_counts_to_show=1,
                                                extensions=("png", "svg"), show_colorbar=True,
                                                bin_number=percent_histogram_bin_number, bin_width=None, bin_array=None,
                                                type="percent", add_max_value=True,
                                                subplot=ax_array[0, 1],
                                                header="#left_xedge\tleft_yedge\tvalue",
                                                save_histovalues_only=True,
                                                logscaled=logscale_heatmaps)

        # Heatmap; x: seq number, y: max_unique_pos_in_seq/aln_len
        print("Heatmap: x: seq number, y: max_unique_pos_in_seq/aln_len")
        MatplotlibRoutines.draw_heatmap(self.general_stats_table[:, 0], 100 * self.general_stats_table[:, 13],
                                        output_prefix="%s.max_unique_pos_seq_number" % output_prefix,
                                        xlabel="N of sequences\nin alignment",
                                        ylabel="Max uniq positions, % of seq", title=None,
                                        figsize=figsize, minimum_counts_to_show=1,
                                        extensions=extensions, show_colorbar=True,
                                        bin_number=(seq_bin_number, percent_histogram_bin_number),
                                        bin_width=None, bin_array=None, min_x_value=min_seq_number_in_alignment,
                                        max_x_value=max_seq_number_in_alignment, min_y_value=0, max_y_value=100,
                                        add_max_value=True, subplot=ax_array[1, 1],
                                        save_histovalues_only=True, header="#left_xedge\tleft_yedge\tcounts",
                                        logscaled=logscale_heatmaps)

        # Heatmap; x: seq number, y: max_unique_insertions_in_seq/aln_len
        print("Heatmap: x: seq number, y: max_unique_insertions_in_seq/aln_len")
        MatplotlibRoutines.draw_heatmap(self.general_stats_table[:, 0], 100 * self.general_stats_table[:, 5],
                                        output_prefix="%s.max_unique_insertions_seq_number" % output_prefix,
                                        xlabel=None, #"N of sequences\nin alignment",
                                        ylabel="Max uniq insertions, % of seq", title=None,
                                        figsize=figsize, minimum_counts_to_show=1,
                                        extensions=extensions, show_colorbar=True,
                                        bin_number=(seq_bin_number, percent_histogram_bin_number),
                                        bin_width=None, bin_array=None, min_x_value=min_seq_number_in_alignment,
                                        max_x_value=max_seq_number_in_alignment, min_y_value=0, max_y_value=100,
                                        add_max_value=True, subplot=ax_array[0, 2],
                                        save_histovalues_only=True, header="#left_xedge\tleft_yedge\tcounts",
                                        logscaled=logscale_heatmaps)

        # Heatmap: x: seq number, y: max_unique_gaps_in_seq/aln_len
        print("Heatmap: x: seq number, y: max_unique_gaps_in_seq/aln_len")
        MatplotlibRoutines.draw_heatmap(self.general_stats_table[:, 0], 100 * self.general_stats_table[:, 7],
                                        output_prefix="%s.max_unique_gaps_seq_number" % output_prefix,
                                        xlabel="N of sequences\nin alignment",
                                        ylabel="Max uniq gaps, % of seq", title=None,
                                        figsize=figsize, minimum_counts_to_show=1,
                                        extensions=extensions, show_colorbar=True,
                                        bin_number=(seq_bin_number, percent_histogram_bin_number),
                                        bin_width=None, bin_array=None, min_x_value=min_seq_number_in_alignment,
                                        max_x_value=max_seq_number_in_alignment, min_y_value=0, max_y_value=100,
                                        add_max_value=True, subplot=ax_array[1, 2],
                                        save_histovalues_only=True, header="#left_xedge\tleft_yedge\tcounts",
                                        logscaled=logscale_heatmaps)


        # Heatmap; x: seq number, y: max_unique_leading_pos_in_seq/aln_len
        print("Heatmap: x: seq number, y: max_unique_leading_pos_in_seq/aln_len")
        MatplotlibRoutines.draw_heatmap(self.general_stats_table[:, 0], 100 * self.general_stats_table[:, 9],
                                        output_prefix="%s.max_unique_leading_pos_seq_number" % output_prefix,
                                        xlabel=None, #"N of sequences\nin alignment",
                                        ylabel="Max uniq leading pos, % of seq", title=None,
                                        figsize=figsize, minimum_counts_to_show=1,
                                        extensions=extensions, show_colorbar=True,
                                        bin_number=(seq_bin_number, percent_histogram_bin_number),
                                        bin_width=None, bin_array=None, min_x_value=min_seq_number_in_alignment,
                                        max_x_value=max_seq_number_in_alignment, min_y_value=0, max_y_value=100,
                                        add_max_value=True, subplot=ax_array[0, 3],
                                        save_histovalues_only=True, header="#left_xedge\tleft_yedge\tcounts",
                                        logscaled=logscale_heatmaps)

        # Heatmap: x: seq number, y: max_unique_trailing_pos_in_seq/aln_len
        print("Heatmap: x: seq number, y: max_unique_trailing_pos_in_seq/aln_len")
        MatplotlibRoutines.draw_heatmap(self.general_stats_table[:, 0], 100 * self.general_stats_table[:, 11],
                                        output_prefix="%s.max_unique_trailing_pos_seq_number" % output_prefix,
                                        xlabel="N of sequences\nin alignment",
                                        ylabel="Max uniq trailing pos, % of seq", title=None,
                                        figsize=figsize, minimum_counts_to_show=1,
                                        extensions=extensions, show_colorbar=True,
                                        bin_number=(seq_bin_number, percent_histogram_bin_number),
                                        bin_width=None, bin_array=None, min_x_value=min_seq_number_in_alignment,
                                        max_x_value=max_seq_number_in_alignment, min_y_value=0, max_y_value=100,
                                        add_max_value=True, subplot=ax_array[1, 3],
                                        save_histovalues_only=True, header="#left_xedge\tleft_yedge\tcounts",
                                        logscaled=logscale_heatmaps)

        plt.tight_layout()
        #aaa = self.general_stats_table[:, 3].astype(float) / self.general_stats_table[:, 1].astype(float) * 100
        #for i in range(0, len(aaa)):
        #    print "%i\t%i\t%i\t%f" % (i, self.general_stats_table[:, 3][i], self.general_stats_table[:, 1][i], aaa[i])

        for ext in extensions:
            plt.savefig("%s.%s" % (output_prefix, ext))

    """
    def __str__(self):
        collection_string = ""
        if self.metadata:
            collection_string += str(self.metadata)
        if self.header:
            collection_string += "\n" + str(self.header)
        if self.records:
            collection_string += "\n" + "\n".join([str(record) for record in self.records])

        return collection_string

    def __iter__(self):
        for record in self.records:
            yield record

    def __getitem__(self, item):
        #string-like access
        return self.records[item]

    def __len__(self):
        return len(self.records)

    def pop(self, index=None):
        # i am a fairy idiot - forgot that 0 is equal to False in python.
        # What to do? Everywhere make direct check if variable is None
        return self.records.pop() if index is None else self.records.pop(index)

    def filter_records_by_parameter_value(self, parameter, minimum=None, maximum=None):
        # TODO: check
        filtered_records = []
        filtered_out_records = []
        if (minimum is not None) and (maximum is not None):
            comparison = lambda par, min_value, max_value: min_value <= par <= max_value
        elif (minimum is None) and (maximum is None):
            raise ValueError("Both minimum and maximum thresholds were not set")
        elif minimum is None:
            comparison = lambda par, min_value, max_value: par <= max_value
        else:
            comparison = lambda par, min_value, max_value: min_value <= par

        for record in self.records:
            if comparison(getattr(record, parameter), minimum, maximum):
                filtered_records.append(record)
            else:
                filtered_out_records.append(record)

        return filtered_records, filtered_out_records

    def filter_records_by_expression(self, expression):
        filtered_records = []
        filtered_out_records = []
        for record in self.records:
            if eval(expression):
                filtered_records.append(record)
            else:
                filtered_out_records.append(record)

        return filtered_records, filtered_out_records

    def filter_records_by_function(self, function):
        filtered_records = []
        filtered_out_records = []
        for record in self.records:
            #print("a\na\na\na\na\na\na\na")
            #print(record.description["Power"])
            if function(record):
                filtered_records.append(record)
            else:
                filtered_out_records.append(record)

        return filtered_records, filtered_out_records

    def check_location(self, bad_region_collection_gff, expression="bad_region.start <= self.pos <= bad_region.end"):
        for record in self:
            record.check_location(bad_region_collection_gff, expression=expression)
    """
    """
    def filter_by_expression(self, expression):
        self_type = self.__class__.__name__
        filtered_records, filtered_out_records = self.filter_records_by_expression(expression)
        #TODO: think how to rewrite following damned string!!!!!!!!!!
        exec("from Parsers.%s import %s" % (self_type[10:], self_type))
        return eval(self_type)(metadata=self.metadata, record_list=filtered_records,
                               header_list=self.header_list, from_file=False), \
               eval(self_type)(metadata=self.metadata, record_list=filtered_out_records,
                               header_list=self.header_list, from_file=False)
    """
    """
    def split_records_by_flags(self, flag_set, mode="all"):
        # possible modes:
        # all - record to be counted as 'with flag' must have all flags from flags_list
        # one - record to be counted as 'with flag' must have at least one flag from flags_list
        flags = set(flag_set)
        with_flags_records = []
        without_flags_records = []
        if mode == "all":
            for record in self:
                if flags & record.flags == flags:
                    with_flags_records.append(record)
                else:
                    without_flags_records.append(record)
        elif mode == "one":
            for record in self:
                if flags & record.flags:
                    with_flags_records.append(record)
                else:
                    without_flags_records.append(record)

        return with_flags_records, without_flags_records

    def write(self, output_file):
        with open(output_file, "w") as out_fd:
            if self.metadata:
                out_fd.write(str(self.metadata) + "\n")
            if self.header:
                out_fd.write(str(self.header) + "\n")
            for record in self.records:
                out_fd.write(str(record) + "\n")

    def get_location(self, record_dict, key="Loc", use_synonym=False, synonym_dict=None):
        for record in self.records:
            record.get_location(record_dict, key=key, use_synonym=use_synonym, synonym_dict=synonym_dict)

    def _split_regions(self):
        splited_dict = OrderedDict({})
        for record in self.records:
            if record.chrom not in splited_dict:
                splited_dict[record.chrom] = [record]
            else:
                splited_dict[record.chrom].append(record)
        return splited_dict
    """



#!/usr/bin/env python
__author__ = 'Sergei F. Kliver'

from collections import OrderedDict
import numpy as np
import pandas as pd
from RouToolPa.Tools.Abstract import Tool
from RouToolPa.Routines import MathRoutines




class GenomeCov(Tool):
    def __init__(self, path="", max_threads=4):
        Tool.__init__(self, "bedtools genomecov", path=path, max_threads=max_threads)

    @staticmethod
    def parse_options(input_bam=None, report_zero_coverage=None,
                      one_based_coordinates=None, scale=None, bedgraph_output=False,
                      genome_bed=None):

        options = " -ibam %s" % input_bam
        if bedgraph_output and report_zero_coverage:
            options += " -bga"
        elif bedgraph_output:
            options += " -bg"
        else:
            options += " -d" if one_based_coordinates else " -dz"
        options += " -scale %f" % scale if scale else ""
        options += " -g %s" % genome_bed if genome_bed else ""

        return options

    def get_coverage(self, output, input_bam=None, report_zero_coverage=None,
                     one_based_coordinates=None, scale=None, bedgraph_output=False,
                     genome_bed=None):

        options = self.parse_options(input_bam=input_bam, report_zero_coverage=report_zero_coverage,
                                     one_based_coordinates=one_based_coordinates, scale=scale,
                                     bedgraph_output=bedgraph_output, genome_bed=genome_bed)
        options += " > %s" % output

        self.execute(options)

    def get_coverage_for_gff(self, input_file, genome_bed, output=None):

        options = " -i %s" % input_file
        options += " -g %s" % genome_bed
        options += " > %s" % output if output else ""

        self.execute(options=options)

    def get_bam_coverage_stats(self, input_bam, output_prefix, genome_bed=None,
                               max_coverage=None, min_coverage=None,
                               verbose=True):
        #options_list = []

        each_position_coverage_file = "%s.tab" % output_prefix
        coverage_stat_file = "%s.stat" % output_prefix

        each_position_options = self.parse_options(input_bam=input_bam, report_zero_coverage=True,
                                                   one_based_coordinates=True, genome_bed=genome_bed)
        each_position_options += " > %s" % each_position_coverage_file

        region_options = self.parse_options(input_bam=input_bam, report_zero_coverage=True,
                                            bedgraph_output=True, genome_bed=genome_bed)

        region_options += " > %s.bedgraph" % output_prefix

        options_list = [each_position_options, region_options]

        self.parallel_execute(options_list, threads=2)

        #~/Soft/MAVR/scripts/math/get_stats_from_numer_data.py -i A_ventralis_pe.nodup.q20.tab -o A_ventralis_pe.nodup.q20.stat -l 2

        MathRoutines.get_stats_from_file(each_position_coverage_file, minimum=min_coverage, maximum=max_coverage,
                                         dtype=float, comments="#", delimiter="\t", converters=None, skiprows=0,
                                         usecols=2, unpack=False, ndmin=0, output_file=coverage_stat_file,
                                         verbose=verbose)
    
    def get_coverage_stats(self, coverage_file, output, verbose=True):
        print("Reading...")
        coverage_df = pd.read_csv(coverage_file, sep='\t', header=None, index_col=(0, 1), 
                                  names=("scaffold", "position", "coverage"))

        print("Calculating stats...")
        stat_df = coverage_df.groupby(level=0).agg(["min", "max", "mean", "median"])

        stat_df.to_csv(output, sep="\t")
        if verbose:
            print(stat_df)
        
        return stat_df
    
    def collapse_coverage_file(self, coverage_file, output_file):

        awk_string = "awk -F'\\t' 'BEGIN {SCAF=\"\"; LEN=\"\"; COV=\"\"} {if (($1 != SCAF)) {if (NR > 1) {printf \"%%s\\t%%s\\t%%s\\n\",SCAF,LEN, COV}; SCAF=$1; LEN=$2; COV=$3} else {LEN=$2; COV=COV\",\"$3}} ; END {printf \"%%s\t%%s\t%%s\n\",SCAF,LEN, COV} %s > %s" % (coverage_file, output_file)

        self.execute(options="", cmd=awk_string)

    def convert_per_base_coverage_file_to_vcf(self, coverage_file, output_vcf, sample_name, reference_file,
                                              parsing_mode="parse", format="fasta"):
        vcf_header = "##fileformat=VCFv4.2\n"
        vcf_header += '##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Approximate read depth">\n'
        vcf_header += "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n" % sample_name

        reference_dict = self.parse_seq_file(reference_file, parsing_mode, format=format)

        with self.metaopen(output_vcf, "w") as out_fd:
            out_fd.write(vcf_header)

            for line_list in self.file_line_as_list_generator(coverage_file):
                out_fd.write("%s\t%s\t.\t%s\t.\t.\t.\t.\tDP\t%s\n" % (line_list[0],
                                                                      line_list[1],
                                                                      str(reference_dict[line_list[0]].seq[int(line_list[1])-1]),
                                                                      line_list[2]))

    @staticmethod
    def extract_data_for_cds_from_collapsed_coverage_file(collapsed_coverage_file, cds_bed_file, output_file,
                                                          skip_transcript_with_no_cds=False):
        cds_dict = OrderedDict()

        with open(cds_bed_file, "r") as cds_fd:
            for line in cds_fd:
                line_list = line.strip().split("\t")
                # convert coordinates to python notation
                cds_dict[line_list[0]] = (int(line_list[1]) - 1, int(line_list[2]))
        with open(collapsed_coverage_file, "r") as col_cov_fd:
            with open(output_file, "w") as out_fd:
                for line in col_cov_fd:
                    transcript_id, length, coverage_array = line.strip().split("\t")

                    if transcript_id not in cds_dict:
                        print("No CDS for transcript %s. %s" % (transcript_id, "Skipping..." if skip_transcript_with_no_cds else ""))
                        if skip_transcript_with_no_cds:
                            continue
                        out_fd.write(line)
                        continue
                    coverage_array = coverage_array.split(",")
                    cds_len = cds_dict[transcript_id][1] - cds_dict[transcript_id][0]
                    out_fd.write("%s\t%s\t%s\n" % (transcript_id, str(cds_len),
                                                   ",".join(coverage_array[cds_dict[transcript_id][0]:cds_dict[transcript_id][1]])))

    @staticmethod
    def analyze_collapsed_coverage_file(collapsed_file, output_file):
        line_number = 0

        def int_through_float(string):
            return int(float(string))

        with open(collapsed_file, "r") as in_fd:
            with open(output_file, "w") as out_fd:

                output_header = "#Scaffold\tLength\tMean_coverage\tMedian_coverage\tMin_coverage\tMax_coverage\tCoveraged_positions\tZero_coverage_position_number\tZero_coveraged_region_number\tLeading_zero_covarage_len\tTrailing_zero_covarage_len\tZero_coverage_region_coordinates\tMean_coverage(without_zerocoveraged_ends)\tMedian_coverage(without_zerocoveraged_ends)\tLength_without_zero_coverage_ends\n"

                out_fd.write(output_header)
                for line in in_fd:
                    line_number += 1
                    #print (line_number)
                    #print [line]
                    if line == "\n" or line == "":    # skip blank lines
                        continue
                    tmp = line.strip().split("\t")
                    #print tmp
                    record_id = tmp[0]
                    record_len = int(tmp[1])

                    try:
                        coverage_array = map(int, tmp[2].split(","))
                    except ValueError:
                        coverage_array = map(int_through_float, tmp[2].split(","))

                    if record_len != len(coverage_array):
                        raise ValueError("Malformed line %i" % line_number)

                    if sum(coverage_array) == 0:
                        continue

                    mean_coverage = float(np.mean(coverage_array))
                    median_coverage = float(np.median(coverage_array))
                    min_coverage = float(np.min(coverage_array))
                    max_coverage = float(np.max(coverage_array))
                    coveraged_position_number = np.count_nonzero(coverage_array)

                    zero_coverage_position_number, zero_coverage_regions_list = MathRoutines.find_flat_regions_in_array(coverage_array, value=0)

                    if zero_coverage_position_number > 0:

                        if zero_coverage_regions_list[0][0] == 0:
                            leading_zero_coverage_len = zero_coverage_regions_list[0][1]
                            start_coverage_coordinate = zero_coverage_regions_list[0][1]
                        else:
                            leading_zero_coverage_len = 0
                            start_coverage_coordinate = 0

                        if zero_coverage_regions_list[-1][0] + zero_coverage_regions_list[-1][1] == record_len:
                            trailing_zero_coverage_len = zero_coverage_regions_list[-1][1]
                            end_coverage_coordinate = zero_coverage_regions_list[-1][0]
                        else:
                            trailing_zero_coverage_len = 0
                            end_coverage_coordinate = record_len

                        zero_coveraged_region_number = len(zero_coverage_regions_list)
                        zero_coverage_coordinates_list = []
                        for (start, length) in zero_coverage_regions_list:
                            zero_coverage_coordinates_list.append("%i-%i" % (start + 1, start + length))

                    else:
                        leading_zero_coverage_len = 0
                        trailing_zero_coverage_len = 0
                        zero_coveraged_region_number = 0
                        start_coverage_coordinate = 0
                        end_coverage_coordinate = record_len
                        zero_coverage_coordinates_list = ["."]

                    length_without_zerocoveraged_ends = end_coverage_coordinate - start_coverage_coordinate
                    mean_coverage_without_zero_coverage_ends = float(np.mean(coverage_array[start_coverage_coordinate:end_coverage_coordinate]))
                    median_coverage_without_zero_coverage_ends = float(np.median(coverage_array[start_coverage_coordinate:end_coverage_coordinate]))

                    out_fd.write("%s\t%i\t%.2f\t%.2f\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%s\t%.2f\t%.2f\t%i\n" % (record_id,
                                                                                                           record_len,
                                                                                                           mean_coverage,
                                                                                                           median_coverage,
                                                                                                           min_coverage,
                                                                                                           max_coverage,
                                                                                                           coveraged_position_number,
                                                                                                           zero_coverage_position_number,
                                                                                                           zero_coveraged_region_number,
                                                                                                           leading_zero_coverage_len,
                                                                                                           trailing_zero_coverage_len,
                                                                                                           ",".join(zero_coverage_coordinates_list),
                                                                                                           mean_coverage_without_zero_coverage_ends,
                                                                                                           median_coverage_without_zero_coverage_ends,
                                                                                                           length_without_zerocoveraged_ends
                                                                                                           ))

    def collapse_bedgraph(self, input_file, output_file):
        with self.metaopen(output_file, "w") as out_fd:
            line_list_generator = self.file_line_as_list_generator(input_file, separator="\t")

            prev_scaffold, prev_start, prev_end = line_list_generator.next()[:3]
            prev_start = int(prev_start)
            prev_end = int(prev_end)

            for line in line_list_generator:

                scaffold = line[0]
                start = int(line[1])
                end = int(line[2])

                if (scaffold != prev_scaffold) or (start != prev_end):
                    out_fd.write("%s\t%i\t%i\n" % (prev_scaffold, prev_start, prev_end))
                    prev_scaffold = scaffold
                    prev_start = start
                    prev_end = end
                else:
                    prev_end = end

            out_fd.write("%s\t%i\t%i\n" % (prev_scaffold, prev_start, prev_end))









#!/usr/bin/env python
"""
Sequence Parser Module
"""
__author__ = 'Sergei F. Kliver'


import re
from copy import deepcopy
from collections import OrderedDict

import numpy as np
import pandas as pd

from RouToolPa.Routines.File import FileRoutines
from RouToolPa.Parsers.GFF import CollectionGFF


class CollectionSequence(FileRoutines):

    def __init__(self, in_file=None, records=None, description=None,
                 external_description=None,
                 format="fasta",
                 parsing_mode="parse", black_list=(), white_list=(),
                 masking=None, masking_file=None, masking_filetype="gff",
                 verbose=False, seq_expression=None, get_stats=False):
        self.formats = ["fasta"]
        self.format = format
        self.parsing_mode = parsing_mode
        self.seq_file = in_file
        self.seq_file_format = format
        self.white_list = white_list
        self.black_list = black_list
        self.seq_file = in_file
        self.description = OrderedDict()

        if in_file:
            self.read(in_file, format=format, parsing_mode=parsing_mode,
                      black_list=black_list,  white_list=white_list, verbose=verbose,
                      seq_expression=seq_expression)

        elif records is None:
            self.records = OrderedDict()
            self.description = description if description else OrderedDict()
        else:
            self.records = records
            self.description = description if description else OrderedDict()

        self.external_description = external_description if external_description else OrderedDict()

        if masking_file:
            print("Parsing masking...")
            self.masking = CollectionGFF(masking_file, format=masking_filetype, parsing_mode="only_coordinates",
                                         black_list=black_list, white_list=white_list)
        else:
            self.masking = masking

        self.seq_lengths = None # None or pandas dataframe with seq_id as index
        self.length = 0
        self.scaffolds = None
        self.gaps = None          # None or pandas dataframe with seq_id as index

        if get_stats:
            self.get_stats_and_features(count_gaps=False, sort=False, min_gap_length=1)

    def sequence_generator(self, sequence_file, format="fasta", black_list=(), white_list=(), verbose=False):

        if format == "fasta":
            with self.metaopen(sequence_file, "r") as seq_fd:
                seq_id = None
                description = None
                seq = ""
                for line in seq_fd:
                    if line[0] == ">":
                        if seq_id and (seq_id not in black_list):
                            if (not white_list) or (seq_id in white_list):
                                if verbose:
                                    print("Parsing %s" % seq_id)
                                yield seq_id, description, seq
                        tmp = line[1:].strip().split(None, 1)
                        seq_id, description = tmp if len(tmp) == 2 else (tmp[0], "")

                        seq = ""
                    else:
                        seq += line[:-1]
                else:
                    if seq_id and (seq_id not in black_list):
                        if (not white_list) or (seq_id in white_list):
                            if verbose:
                                print("Parsing %s" % seq_id)
                            yield seq_id, description, seq

    def sequence_generator_with_expression(self, sequence_file, seq_expression, format="fasta", black_list=(), white_list=(), verbose=False,
                           ):

        if format == "fasta":
            with self.metaopen(sequence_file, "r") as seq_fd:
                seq_id = None
                description = None
                seq = ""
                for line in seq_fd:
                    if line[0] == ">":
                        if seq_id and (seq_id not in black_list):
                            if (not white_list) or (seq_id in white_list):
                                if verbose:
                                    print("Parsing %s" % seq_id)
                                yield seq_id, description, seq
                        tmp = line[1:].strip().split(None, 1)
                        seq_id, description = tmp if len(tmp) == 2 else (tmp[0], "")

                        seq = ""
                    else:
                        seq += line[:-1]
                else:
                    if seq_id and (seq_id not in black_list):
                        if (not white_list) or (seq_id in white_list):
                            if verbose:
                                print("Parsing %s" % seq_id)
                            yield seq_id, description, seq_expression(seq)

    def sequence_tuple_generator(self):
        if self.parsing_mode == "parse":
            for scaffold_id in self.scaffolds:
                yield (scaffold_id, self.records[scaffold_id])

    def reset_seq_generator(self):
        self.records = self.sequence_generator(self.seq_file, format=self.seq_file_format,
                                               black_list=self.black_list,  white_list=self.white_list)

    def read(self, seq_file, format="fasta", parsing_mode="generator", black_list=(), white_list=(),
             verbose=False, seq_expression=None):
        if format not in self.formats:
            raise ValueError("ERROR!!! This format(%s) was not implemented yet for parsing!" % parsing_mode)
        if parsing_mode == "generator":
            print("Creating sequence generator...")
            if seq_expression:
                self.records = self.sequence_generator_with_expression(seq_file,
                                                                       seq_expression,
                                                                       format=format,
                                                                       black_list=black_list,
                                                                       white_list=white_list,
                                                                       verbose=verbose)
            else:
                self.records = self.sequence_generator(seq_file,
                                                       format=format,
                                                       black_list=black_list,
                                                       white_list=white_list,
                                                       verbose=verbose)
        elif parsing_mode == "parse":
            print("Parsing sequences...")
            self.records = OrderedDict()
            for seq_id, description, seq in self.sequence_generator(seq_file, format=format,
                                                                    black_list=black_list,
                                                                    white_list=white_list):
                self.records[seq_id] = seq
                self.description[seq_id] = description

            return self.records

    def __len__(self):
        """
        :return: length of genome or None if genome was not parsed yet
        """
        return self.length

    def scaffold_number(self):
        """
        :return: number of regions/scaffolds/chromosomes in genome
        """
        return len(self.scaffolds)

    def get_stats_and_features(self, count_gaps=True, sort=True, min_gap_length=1):
        length_list = []
        gaps_list = []
        if self.parsing_mode == "generator":
            for seq_id, description, seq in self.records:
                length_list.append([seq_id, len(seq)])
                if count_gaps:
                    gaps_list.append(self.find_gaps_in_seq(seq, seq_id, min_gap_length=min_gap_length))

            self.reset_seq_generator()
        else:
            for seq_id in self.records:
                length_list.append([seq_id, len(self.records[seq_id])])
                if count_gaps:
                    gaps_list.append(self.find_gaps_in_seq(self.records[seq_id], seq_id, min_gap_length=min_gap_length))
        if count_gaps:
            self.gaps = CollectionGFF(records=pd.concat(gaps_list), format="gff", parsing_mode="only_coordinates",
                                      black_list=self.black_list, white_list=self.white_list
                                      )

            self.gaps.records.sort_values(by=["scaffold", "start", "end"])

        self.seq_lengths = pd.DataFrame.from_records(length_list, columns=("scaffold", "length"), index="scaffold")
        if sort:
            self.seq_lengths.sort_values(by=["length", "scaffold"])
        self.length = np.sum(self.seq_lengths["length"])
        self.scaffolds = self.seq_lengths.index.values

    @staticmethod
    def find_gaps_in_seq(sequence, seq_id=None, min_gap_length=1):
        """
        Finds gaps (N) in sequence
        :return: None
        """
        gap_reg_exp = re.compile("N+", re.IGNORECASE)
        gaps_list = []
        gaps = gap_reg_exp.finditer(sequence)  # iterator with
        for match in gaps:
            if (match.end() - match.start()) >= min_gap_length:
                gaps_list.append([match.start(), match.end()])
        gaps_list = pd.DataFrame(gaps_list, columns=("start", "end"))
        if seq_id:
            # add seq id as index
            gaps_list.index = pd.Index([seq_id] * len(gaps_list), name="scaffold")
        return gaps_list

    def remove_small_gaps(self, min_gap_length):
        self.gaps = self.gaps[(self.gaps['end'] - self.gaps['start']) >= min_gap_length]

    def get_merged_gaps_and_masking(self):
        if self.masking and self.gaps:
            merged = self.masking + self.gaps
            merged.collapse_records(sort=False, verbose=True) # sorting is called during addition
            return merged
        elif self.masking:
            return self.masking
        elif self.gaps:
            return self.gaps
        else:
            return None

    def write(self, outfile, expression=None, max_symbols_per_line=60):
        if self.parsing_mode == "parse":
            with self.metaopen(outfile, "w") as out_fd:
                for seq_id in self.records:
                    if expression:
                        if not expression(seq_id, self.records[seq_id]):
                            continue
                    out_fd.write(">%s\n" % seq_id if seq_id not in self.description else ">%s %s\n" % (seq_id, self.description[seq_id]))
                    length = self.seq_lengths[seq_id][0] if self.seq_lengths else len(self.records[seq_id])
                    line_number = length // max_symbols_per_line
                    index = 0
                    while index < line_number:
                        out_fd.write(self.records[seq_id][index*max_symbols_per_line:(index+1)*max_symbols_per_line] + "\n")
                        index += 1
                    if line_number * max_symbols_per_line < length:
                        out_fd.write(self.records[seq_id][index*max_symbols_per_line:] + "\n")

        else:
            raise ValueError("ERROR!!! Writing was implemented only for parsing mode yet!")

    def write_by_syn(self, outfile, syn_dict, max_symbols_per_line=60, absent_symbol="."):
        """
        :param outfile: output fasta file
        :param syn_dict: dictionary with synonyms to be used for sequence ids.
               Note: sequence form collection will be written as many times as it appears in values of syn_dict
        :param max_symbols_per_line: Maximum symbols per line in output fasta
        :return:
        """
        if self.parsing_mode == "parse":
            with self.metaopen(outfile, "w") as out_fd:
                for seq_id in syn_dict:
                    if syn_dict[seq_id] == absent_symbol:
                        continue
                    #out_fd.write(">%s\n" % seq_id)

                    out_fd.write(">%s\n" % seq_id if syn_dict[seq_id] not in self.description else ">%s %s\n" % (seq_id, self.description[syn_dict[seq_id]]))

                    length = self.seq_lengths[syn_dict[seq_id]][0] if self.seq_lengths else len(self.records[syn_dict[seq_id]])
                    line_number = length // max_symbols_per_line
                    index = 0
                    while index < line_number:
                        out_fd.write(self.records[syn_dict[seq_id]][index*max_symbols_per_line:(index+1)*max_symbols_per_line] + "\n")
                        index += 1
                    if line_number * max_symbols_per_line < length:
                        out_fd.write(self.records[syn_dict[seq_id]][index*max_symbols_per_line:] + "\n")

        else:
            raise ValueError("ERROR!!! Writing was implemented only for parsing mode yet!")

    def convert_rna_to_dna(self):
        if self.parsing_mode == "parse":
            for seq_id in self.scaffolds:
                self.records[seq_id] = self.records[seq_id].replace("U", "T").replace("u", "t")
        else:
            raise ValueError("ERROR!!! Convertation from RNA to DNA was implemented only for parsing mode yet!")

    @staticmethod
    def count_window_number_in_scaffold(scaffold_length, window_size, window_step):
        if scaffold_length < window_size:
            return 0
        return int((scaffold_length - window_size)/window_step) + 1

    def count_window_number(self, window_size, window_step):

        def count(scaffold_length):
            return self.count_window_number_in_scaffold(scaffold_length, window_size, window_step)

        return self.length.apply(count)

    @staticmethod
    def expression_unmask(s):
        return s.upper()

    def unmask(self, in_place=False, verbose=None):

        if in_place:
            collection = self
        else:
            collection = deepcopy(self)

        if collection.parsing_mode == "generator":
            print("Creating unmasked sequence generator...")

            collection.records = collection.sequence_generator_with_expression(collection.seq_file,
                                                                               collection.expression_unmask,
                                                                               format=collection.format,
                                                                               black_list=collection.black_list,
                                                                               white_list=collection.white_list,
                                                                               verbose=verbose)
        elif collection.parsing_mode == "parse":
            print("Unmasking sequences...")
            for seq_id in collection.records:
                collection.records[seq_id] = collection.records[seq_id].upper()

        if not in_place:
            return collection



#!/usr/bin/env python
__author__ = 'Sergei F. Kliver'

from collections import Iterable, OrderedDict
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import colors
from RouToolPa.Parsers.Abstract import Record, Collection, Metadata, Header
from RouToolPa.Parsers.VCF import CollectionVCF, MetadataVCF, HeaderVCF

from RouToolPa.General.GeneralCollections import WDict


class RecordCCF(Record, Iterable):

    def __init__(self, size=None, start=None, end=None, info_dict=None, flags=None,
                 collection_vcf=None, bad_vcf_records=0, from_records=True, subclusters=None, features=None,
                 ):
        # possible flags:
        # IP - indel(s) are present in record
        # BR - record is located in bad region
        self.records = collection_vcf
        self.subclusters = subclusters
        if from_records:
            self.size = len(collection_vcf)
            self.start = collection_vcf[0].pos
            self.end = collection_vcf[-1].pos - 1 + \
                       max(map(lambda x: len(x), collection_vcf[-1].alt_list + [collection_vcf[-1].ref]))
            self.flags = set(flags) if flags else set()
            for record in self.records:
                if record.check_indel():
                    self.flags.add("IP")
                    break
            self.info_dict = info_dict if info_dict else OrderedDict({})
            self.features = features

            self.bad_records = 0
            for record in self.records:
                if "BR" in record.flags:
                    self.bad_records += 1
        else:
            self.size = size
            self.start = start
            self.end = end
            self.features = features
            self.info_dict = info_dict
            self.mean_dist = None
            self.flags = set(flags) if flags else set()
            self.bad_records = bad_vcf_records
        self.len = self.end - self.start + 1
        if bad_vcf_records > 0:
            self.flags.add("BR")
        else:
            self.flags - set(["BR"])

        distances = self.distances()
        if self.info_dict is None:
            self.info_dict = OrderedDict({})

        if self.size > 1:
            self.info_dict["Mean"] = np.mean(distances)
            self.info_dict["Median"] = np.median(distances)
            self.info_dict["Power"] = self.size / np.median(distances)

    def __len__(self):
        return self.size

    def __iter__(self):
        for record in self.records:
            yield record

    def __str__(self):
        attributes_string = "Size=%i;Bad_records=%i" % (self.size, self.bad_records)
        if self.flags:
            attributes_string += ";" + ";".join(self.flags)
        if self.info_dict:
            attributes_string += ";" + ";".join(["%s=%s" % (key, ",".join(map(lambda x: str(x), self.info_dict[key])) if isinstance(self.info_dict[key], list) or isinstance(self.info_dict[key], set)  else str(self.info_dict[key]))
                                                 for key in self.info_dict])
        if self.subclusters is not None:
            #print(self.subclusters)
            attributes_string += ";Subclusters=" + ",".join(map(lambda x: str(x), self.subclusters))

        cluster_string = "%i\t%i\t%s" % (self.start, self.end, attributes_string)
        return cluster_string + "\n\t" + "\n\t".join([str(record) for record in self.records])

    def distances(self):
        #returns distances between variants in cluster
        positions = np.array([record.pos for record in self])
        return np.ediff1d(positions)

    def reset_flags(self, flags_to_reset='all'):
        # flags_to_reset can take values 'all', set or list, or None
        # if all - all flags will be removed
        # if None - nothing will be done
        # if set/list - only flags in set/list will be removed
        if flags_to_reset == 'all':
            self.flags = set([])
        elif flags_to_reset:
            flags_to_remove = set(flags_to_reset)
            self.flags = self.flags - flags_to_remove

    def check_flags(self, flag_list, mismatch_list=[], expression_list=[], remove_mismatch_list=[],
                    flags_to_reset=None, mode="all", min_cluster_size=[]):
        # TODO: fix changes in  loc in info_dict after removal
        # at moment subclustering resets after check if some record was removed

        # allow_flag_mismatch - number of allowed mismaches. if zero or None no mismatches are allowed
        self.reset_flags(flags_to_reset)
        if self.flags is None:
            self.flags = set([])


        mismatches = mismatch_list if mismatch_list else [0 for flag in flag_list]
        min_cluster = min_cluster_size if min_cluster_size else [3 for flag in flag_list]
        remove_list = remove_mismatch_list if remove_mismatch_list else[False for flag in flag_list]
        # possible from_records_flag_mode:
        # all - record to be counted as 'with flag' must have all flags from flags_list
        # one - record to be counted as 'with flag' must have at least one flag from flags_list

        # option flags_to_consider works only for remove_mismatch_records
        if mode == "one":
            for record in self:
                self.flags |= record.flags
        elif mode == "all":
            record_to_remove_dict = dict([(flag, set([])) for flag in flag_list])
            mismatch_count_dict = dict([(flag, 0) for flag in flag_list])
            index = 0
            for record in self:
                expressions = [eval(expression) for expression in expression_list] \
                    if expression_list else [True for flag in flag_list]
                for flag, expression, min_size in zip(flag_list, expressions, min_cluster):
                    #if self.size < min_size:
                    #    continue
                    if flag not in record.flags:
                        if mismatch_count_dict[flag] is not None:
                            mismatch_count_dict[flag] += 1
                        if expression and self.size >= min_size:# expression to reconsider cluster with mismatch as good
                            record_to_remove_dict[flag].add(index)
                        else:
                            # if expression is not True flag will not be set, None is used to indicate it
                            mismatch_count_dict[flag] = None
                index += 1
            for flag, mismatch, min_size in zip(flag_list, mismatches, min_cluster):
                if (mismatch_count_dict[flag] is not None) and mismatch_count_dict[flag] <= mismatch:

                    self.flags.add(flag)
                    if mismatch_count_dict[flag] > 0:
                        self.info_dict["N%sR" % flag] = [mismatch_count_dict[flag]]
                else:
                    record_to_remove_dict[flag].clear()
            records_to_remove = set([])
            for flag, remove in zip(flag_list, remove_list):
                if remove and (mismatch_count_dict[flag] is not None):
                    records_to_remove |= record_to_remove_dict[flag]
            records_to_remove = sorted(records_to_remove, reverse=True)
            #print(records_to_remove)
            record_scaffold = self.records.scaffold_list[0]
            if records_to_remove:
                #print(self)
                for index in records_to_remove:
                    self.records.records[record_scaffold].pop(index)
                    self.subclusters = np.delete(self.subclusters, index)
                self.__init__(collection_vcf=CollectionVCF(metadata=self.records.metadata,
                                                           records_dict=self.records.records,
                                                           header=self.records.header,
                                                           samples=self.records.samples, from_file=False),
                              info_dict=self.info_dict,
                              flags=self.flags, subclusters=self.subclusters,
                              from_records=True)

    def check_location(self):
        self.bad_records = 0
        for variant in self:
            if "BR" in variant.flags:
                self.bad_records += 1
        if self.bad_records > 0:
            self.flags.add("BR")

    def get_location(self, record_scaffold, record_dict, key="Loc", use_synonym=False, synonym_dict=None,
                     feature_type_black_list=[], strand_key=None):
        # function is written for old variant (with sub_feature)s rather then new (with CompoundLocation)
        # id of one SeqRecord in record_dict must be equal to record.pos
        if not self.info_dict:
            self.info_dict = {}
        if not self.features:
            self.features = []
        if key not in self.info_dict:
            self.info_dict[key] = set([])
        for variant in self:
            if key in variant.info_dict:
                self.info_dict[key] |= set(variant.info_dict[key])
            for feature in record_dict[record_scaffold].features:
                if (variant.pos - 1) in feature:
                    self.features.append(feature)
                    self.info_dict[key].add(self.get_synonym(feature.type, use_synonym=use_synonym,
                                                               synonym_dict=synonym_dict))
                for sub_feature in feature.sub_features:
                    if (variant.pos - 1) in sub_feature:
                        self.info_dict[key].add(self.get_synonym(sub_feature.type, use_synonym=use_synonym,
                                                                   synonym_dict=synonym_dict))

    def subclustering(self,
                      method="inconsistent",
                      threshold=0.8,
                      cluster_distance='average'):
        tmp = self.records.get_clusters(extracting_method=method,
                                        threshold=threshold,
                                        cluster_distance=cluster_distance,
                                        draw_dendrogramm=False,
                                        return_collection=False,
                                        write_inconsistent=False,
                                        write_correlation=False)
        self.subclusters = tmp[tmp.keys()[0]]

    def adjust(self, border_limit=None, min_size_to_adjust=2, remove_border_subclusters=False, remove_size_limit=1):
        # adjusts cluster borders, returns list of new cluster records
        # skip adjustment for clusters with 3 or less mutations
        if (self.size < min_size_to_adjust) or (self.subclusters is None):
            #return -1
            return [self]
        limit = border_limit if border_limit else len(self.subclusters)
        for i in range(0, limit):
            if self.subclusters[i] == self.subclusters[0]:
                left_subcluster_end = i
            else:
                break
        # exit if cluster doesnt have subclusters
        if left_subcluster_end == len(self.subclusters) - 1:
            #return 1
            return [self]

        for i in range(-1, -limit - 1, -1):
            if self.subclusters[i] == self.subclusters[-1]:
                right_subcluster_start = i
            else:
                break

        if remove_border_subclusters:
            start = left_subcluster_end + 1 if left_subcluster_end < remove_size_limit else 0
            end = right_subcluster_start if right_subcluster_start >= -remove_size_limit else len(self.subclusters)

            new_left_cluster, new_right_cluster = None, None
            record_scaffold = self.records.scaffold_list[0]

            if start > 0:
                new_left_cluster = RecordCCF(collection_vcf=CollectionVCF(records_dict=dict([(record_scaffold, self.records.records[record_scaffold][:start])]),
                                                                          from_file=False),
                                             subclusters=self.subclusters[:start], from_records=True)

            if end < len(self.subclusters):
                new_right_cluster = RecordCCF(collection_vcf=CollectionVCF(records_dict=dict([(record_scaffold, self.records.records[record_scaffold][end:])]),
                                                                           from_file=False),
                                              subclusters=self.subclusters[end:], from_records=True)

            new_middle_cluster = RecordCCF(collection_vcf=CollectionVCF(records_dict=dict([(record_scaffold, self.records.records[record_scaffold][start:end])]),
                                                                        from_file=False),
                                           subclusters=self.subclusters[start:end], from_records=True)

            cluster_list = [new_left_cluster] if new_left_cluster else []
            cluster_list += [new_middle_cluster]
            cluster_list += [new_right_cluster] if new_right_cluster else []
            return cluster_list

    def check_strandness(self):
        #for desaminases only
        count_C = 0.
        for record in self:
            if record.ref == "C":
                count_C += 1.
        homogeneity = count_C / self.size
        self.info_dict["Homogeneity"] = homogeneity if homogeneity >= 0.5 else 1.0 - homogeneity
        self.info_dict["Strand"] = "P" if homogeneity >= 0.5 else "N"

    def gff_string(self):
        attributes_string = "Size=%i;Bad_records=%i" % (self.size, self.bad_records)
        if self.info_dict:
            for key in self.info_dict:
                attributes_string += ";%s=%s" % (key, self.info_dict[key])
        if self.flags:
            attributes_string += ";" + ";".join(self.flags)
        if self.subclusters != None:
            #print(self.subclusters)
            attributes_string += ";Subclusters=" + ",".join(map(lambda x: str(x), self.subclusters))
        return "%s\t%s\t%i\t%i\t.\t.\t.\t%s" % ("custom", "cluster", self.start, self.end, attributes_string)


class MetadataCCF(Metadata):

    def __init__(self, samples, vcf_metadata=None, vcf_header=None, metadata={}):
        self.samples = samples      #list
        self.metadata = metadata
        self.vcf_metadata = vcf_metadata
        self.vcf_header = vcf_header

    def __str__(self):
        metadata_string = None
        if self.vcf_metadata:
            metadata_string = "#VCF_METADATA START\n" + str(self.vcf_metadata) + "\n#VCF_METADATA END"
        if self.vcf_header:
            metadata_string = metadata_string + "\n#VCF_HEADER\n" + str(self.vcf_header) \
                if metadata_string else "#VCF_HEADER\n" + str(self.vcf_header)
        if self.metadata:
            metadata_string += "\n##" + "\n##".join(["%s=%s" % (key, self.metadata[key]) for key in self.metadata])
        return metadata_string


class HeaderCCF(list, Header):
    def __str__(self):
        return "#CCF_HEADER\n#" + "\t".join(self)


class CollectionCCF(Collection):
    """
    def read(self, input_file):
        # TODO: write read from ccf file
        with open(input_file, "r") as in_fd:

                stripped_line = in_fd.readline().strip()
                if stripped_line == "#VCF_METADATA START":
                    vcf_metadata = MetadataVCF()
                    stripped_line = in_fd.readline().strip()
                    while (stripped_line != "#VCF_METADATA END"):
                        vcf_metadata.add_metadata(stripped_line)
                        stripped_line = in_fd.readline().strip()
                stripped_line = in_fd.readline().strip()
                if stripped_line == "#VCF_HEADER":
                    header_line = in_fd.readline().strip()
                    vcf_header = HeaderVCF(header_line[1:].split("\t"))
                    #print("a\na\na\na\na\n")
                    #print(vcf_header)
                    self.metadata = MetadataCCF(vcf_header[9:], vcf_metadata=vcf_metadata, vcf_header=vcf_header)
                stripped_line = in_fd.readline().strip()
                if stripped_line == "#CCF_HEADER":
                    header_line = in_fd.readline().strip()
                    self.header = HeaderCCF(header_line[1:].split("\t"))
                flag = 0
                self.records = []
                while True:
                    data_line = in_fd.readline()

                    if data_line == "" or data_line == "\n":
                        break
                    stripped_line = data_line.strip()
                    if data_line[0] == "\t":
                        #stripped_line = stripped_line[1:]
                        #print(collection_vcf)
                        collection_vcf.records.append(collection_vcf.add_record(stripped_line, external_metadata=self.metadata.vcf_metadata))
                        flag = 1
                        #print("aaaa")
                        continue

                    if flag != 0:
                        self.records.append(RecordCCF(id=cluster_id, chrom=chrom, size=size, start=start, end=end,
                                                      info_dict=info_dict, flags=flags,
                                                      collection_vcf=collection_vcf, bad_vcf_records=bad_records,
                                                      from_records=False, subclusters=subclusters))
                        #collection_vcf = None

                    if stripped_line[0] == ">":
                        flag = 0
                        cluster_id, chrom, start, end, info_dict_and_flags = stripped_line[1:].split("\t")
                        start = int(start)
                        end = int(end)
                        info_dict_and_flags = info_dict_and_flags.split(";")
                        info_dict = OrderedDict({})
                        flags = set([])
                        subclusters = None
                        for descr_entry in info_dict_and_flags:
                            descr_entry_splited = descr_entry.split("=")
                            if len(descr_entry_splited) == 1:
                                flags.add(descr_entry_splited[0])
                                continue
                            if descr_entry_splited[0] == "Size":
                                size = int(descr_entry_splited[1])
                            elif descr_entry_splited[0] == "Bad_records":
                                bad_records = int(descr_entry_splited[1])
                            elif descr_entry_splited[0] == "Mean" or descr_entry_splited[0] == "Median" or descr_entry_splited[0] == "Power" or descr_entry_splited[0] == "Homogeneity":
                                info_dict[descr_entry_splited[0]] = float(descr_entry_splited[1])
                            elif descr_entry_splited[0] == "Loc":
                                info_dict[descr_entry_splited[0]] = descr_entry_splited[1].split(",")
                            elif descr_entry_splited[0] == "Subclusters":
                                subclusters = [int(x) for x in descr_entry_splited[1].split(",")]
                            else:
                                info_dict[descr_entry_splited[0]] = descr_entry_splited[1].split(",")
                                if len(info_dict[descr_entry_splited[0]]) == 1:
                                    info_dict[descr_entry_splited[0]] = info_dict[descr_entry_splited[0]][0]
                        collection_vcf = CollectionVCF(metadata=None, record_list=None, header=None, vcf_file=None, samples=None, from_file=False, external_metadata=None)
                        continue
                self.records.append(RecordCCF(id=cluster_id, chrom=chrom, size=size, start=start, end=end,
                                              info_dict=info_dict, flags=flags,
                                              collection_vcf=collection_vcf, bad_vcf_records=bad_records,
                                              from_records=False, subclusters=subclusters))
    """
    def filter(self, expression):
        # expression should be a function with one argument - record entry
        filtered_records, filtered_out_records = self.filter_records(expression)
        return CollectionCCF(metadata=self.metadata, records_dict=filtered_records,
                             from_file=False,
                             header=self.header), \
               CollectionCCF(metadata=self.metadata, records_dict=filtered_out_records,
                             from_file=False,
                             header=self.header)

    def filter_by_size(self, min_size=3):
        return self.filter(lambda record: record.size >= min_size)

    def count(self):
        sizes = []
        for record in self:
            sizes.append(record.size)
        return sizes

    def statistics(self, filename="cluster_size_distribution.svg", title="Distribution of sizes of clusters",
                   dpi=150, figsize=(8, 8), facecolor="green", draw_hist=True):
        plt.figure(1, dpi=dpi, figsize=figsize)
        plt.subplot(1, 1, 1)
        plt.suptitle(title)
        counts = self.count()
        if not counts:
            return -1
        maximum = max(counts)
        #print(maximum)
        bins = np.linspace(1, maximum+1, maximum+1)
        #print(bins)
        n, bins, patches = plt.hist(counts, bins, facecolor=facecolor, align='left')
        if draw_hist:
            plt.xticks(np.arange(1, maximum+1, 1))
            plt.xlim(xmin=0.5, xmax=maximum+0.5)
            plt.xlabel("Size")
            plt.ylabel("Number")
            plt.savefig(filename)
        plt.close()
        return WDict([map(int, pair) for pair in zip(bins, n)])

    def filter_by_flags(self, white_flag_list=[], black_flag_list=[]):
        white_list = set(white_flag_list)
        black_list = set(black_flag_list)

        if white_list and not black_list:
            expression = lambda record: white_list & record.flags
        elif black_list and not white_list:
            expression = lambda record: not (black_list & record.flags)
        elif black_list and not white_list:
            expression = lambda record: (white_list & record.flags) and not (black_list & record.flags)
        elif not black_list and (not white_list):
            raise ValueError("Both white and black lists were not set")
        return self.filter(expression)

    def subclustering(self,
                      method="inconsistent",
                      threshold=0.8,
                      cluster_distance='average'):
        for record in self:
            if len(record) < 3:
                continue
            record.subclustering(method=method,
                                 threshold=threshold,
                                 cluster_distance=cluster_distance)

    def check_location(self):
        for record in self:
            record.check_location()

    def check_flags(self, flag_list, mismatch_list=[], expression_list=[], remove_mismatch_list=[],
                    flags_to_reset=None, mode="all", min_cluster_size=[]):
        for record in self:
            record.check_flags(flag_list, mismatch_list=mismatch_list, expression_list=expression_list,
                               remove_mismatch_list=remove_mismatch_list, flags_to_reset=flags_to_reset, mode=mode,
                               min_cluster_size=min_cluster_size)

    def check_strandness(self):
        for record in self:
            record.check_strandness()

    def get_data_for_stat(self, additional_data=("Median", "Mean", "Power")):
        data = []
        for record in self:
            if record.size == 1:
                continue
            data.append([record.len, record.size] + ([record.info_dict[add_data] for add_data in additional_data] if additional_data else []))
        return np.array(data)

    def heatmap_statistics(self, filename="heatmap_statistics.svg", suptitle="Heatmap_statistics",
                           dpi=150, facecolor="green", n_bins_default=20,
                           additional_data=("Median", "Mean", "Power")):
        labels_dict = {"Median": "Median distance",
                       "Mean": "Mean distance",
                       "Power": "Power(Size/median dist.)",
                       "Homogeneity": "Strand specificity"}
        names = ("Length", "Size")
        if additional_data is not None:
            names += additional_data
        data = [self.get_data_for_stat(additional_data=additional_data)[:, i]
                for i in range(0, 2 if additional_data is None else 2 + len(additional_data))]
        data.append(data[1] / data[2])
        data.append((data[1]**2) / data[2])
        size = 6 * len(names)
        plt.figure(1, dpi=dpi, figsize=(size, size))
        plt.suptitle(suptitle)

        if len(data[0]) == 0:
            return 0

        for i in range(0, len(names)):
            for j in range(0, len(names)):
                if i == j:
                    continue
                plt.subplot(len(names), len(names), i * len(names) + j + 1)
                plt.xlabel(labels_dict[names[i]] if names[i] in labels_dict else names[i])
                plt.ylabel(labels_dict[names[j]] if names[j] in labels_dict else names[j])
                n_x_bins = 10 if names[i] == "Homogeneity" else int(max(data[i])) if names[i] == "Size" else \
                    int(max(data[i]) * 20) if names[i] == "Power" else n_bins_default
                n_y_bins = 10 if names[j] == "Homogeneity" else int(max(data[j])) if names[j] == "Size" else \
                    int(max(data[j]) * 20) if names[j] == "Power" else n_bins_default
                #print(names[i])
                #print(data[i])
                #print(names[j])
                #print(data[j])
                #print(n_x_bins, n_y_bins)

                #cmap = colors.ListedColormap(['white', 'red'])
                #bounds = [0, 5, 10]
                #norm = colors.BoundaryNorm(bounds, cmap.N)

                plt.hist2d(data[i], data[j], (n_x_bins, n_y_bins), cmin=1)  # normed=True)
                plt.colorbar()
        plt.savefig(filename)
        plt.close()

    def get_data_for_strand_stat(self):
        num_of_P_clusters = 0
        num_of_N_clusters = 0
        strandness_P_clusters = []
        strandness_N_clusters = []
        size_P_clusters = []
        size_N_clusters = []
        length_P_clusters = []
        length_N_clusters = []

        for record in self:
            if "Strand" not in record.info_dict:
                print(record)
            if record.info_dict["Strand"] == "P":
                num_of_P_clusters += 1
                strandness_P_clusters.append(record.info_dict["SHom"])
                size_P_clusters.append(record.size)
                length_P_clusters.append(record.len)
            else:
                num_of_N_clusters += 1
                strandness_N_clusters.append(record.info_dict["SHom"])
                size_N_clusters.append(record.size)
                length_N_clusters.append(record.len)

        return num_of_P_clusters, strandness_P_clusters, size_P_clusters, length_P_clusters, \
               num_of_N_clusters,  strandness_N_clusters, size_N_clusters, length_N_clusters,

    def strandness_statistics(self, filename="Strandness_statistics.svg", suptitle="Cluster strandness_statistics",
                   dpi=150, figsize=(20, 20), facecolor="green"):
        num_of_P_clusters, strandness_P_clusters, size_P_clusters, length_P_clusters, \
        num_of_N_clusters,  strandness_N_clusters, size_N_clusters, length_N_clusters = \
                                                                                     self.get_data_for_strand_stat()

        plt.figure(1, dpi=dpi, figsize=figsize)
        plt.subplot(3, 4, 1)
        points = np.arange(2)
        bar_width = 0.55

        rects1 = plt.bar(points, [num_of_P_clusters, num_of_N_clusters], bar_width,
                         color='b')

        plt.xlabel('Strandness')
        plt.ylabel('Counts')
        plt.title("Distribution of clusters in strands")
        plt.xticks(points + bar_width, ('P', 'M'))

        for subplot_index, counts, title in zip([5, 9], [strandness_P_clusters, strandness_N_clusters], ["P", "N"]):
            plt.subplot(3, 4, subplot_index)
            if len(counts) == 0:
                continue
            bins = np.linspace(0.5, 1.0, 11)
            plt.hist(counts, bins, facecolor=facecolor)
            plt.xticks(np.arange(0.5, 1.0, 0.1))
            plt.title("Strandness coefficient of clusters in %s strand" % title)

        for subplot_index, coeff, size, title in zip([6, 10], [strandness_P_clusters, strandness_N_clusters],
                                                     [size_P_clusters, size_N_clusters], ["P", "N"]):
            plt.subplot(3, 4, subplot_index)
            if len(coeff) == 0:
                continue
            #plt.plot(size, coeff, "b.")
            #heatmap, xedges, yedges = np.histogram2d(size, coeff, bins=(10, max(size)))
            #extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
            #plt.imshow(heatmap)#, extent=extent)
            plt.hist2d(size, coeff, (max(size), 10))
            #plt.yticks(np.arange(0.5, 1.0, 0.1))
            plt.title("Coefficient and size of clusters in %s strand" % title)
            plt.xlabel("Size of cluster")
            plt.ylabel("Strandness")

        for subplot_index, coeff, length, title in zip([7, 11], [strandness_P_clusters, strandness_N_clusters],
                                                     [length_P_clusters, length_N_clusters], ["P", "N"]):
            plt.subplot(3, 4, subplot_index)
            if len(coeff) == 0:
                continue
            plt.plot(length, coeff, "b.")
            plt.title("Coefficient and length of clusters in %s strand" % title)
            plt.xlabel("Length of cluster")
            plt.ylabel("Strandness")

        for subplot_index, size, length, title in zip([8, 12], [size_P_clusters, size_N_clusters],
                                                     [length_P_clusters, length_N_clusters], ["P", "N"]):
            plt.subplot(3, 4, subplot_index)
            if len(size) == 0:
                continue
            plt.plot(length, size, "b.")
            plt.title("Length and size of clusters in %s strand" % title)
            plt.xlabel("Length of cluster")
            plt.ylabel("Size of cluster")

        plt.suptitle(suptitle)
        plt.savefig(filename)
        plt.close()

    def adjust(self, border_limit=None, min_size_to_adjust=2, remove_border_subclusters=False, remove_size_limit=1):
        new_records_dict = OrderedDict()
        for scaffold in self.scaffold_list:
            new_records_dict[scaffold] = []
        for scaffold in self.scaffold_list:
            for record in self.records[scaffold]:
                new_records_dict[scaffold] += record.adjust(border_limit=border_limit,
                                                            min_size_to_adjust=min_size_to_adjust,
                                                            remove_border_subclusters=remove_border_subclusters,
                                                            remove_size_limit=remove_size_limit)
        self.records = new_records_dict

    def extract_vcf(self):
        vcf_records_dict = OrderedDict()
        for scaffold in self.records:
            vcf_records_dict[scaffold] = []
            for record_ccf in self.records[scaffold]:
                vcf_records_dict[scaffold] += record_ccf.records.records[scaffold]
        return CollectionVCF(metadata=self.metadata.vcf_metadata,
                             records_dict=vcf_records_dict,
                             header=self.metadata.vcf_header,
                             samples=self.metadata.samples, from_file=False)

    def write_gff(self, outfile):
        with open(outfile, "w") as out_fd:
            for scaffold in self.scaffold_list:
                for record in self.records[scaffold]:
                    out_fd.write("%s\t%s\n" % (scaffold, record.gff_string()))